#include <stdlib.h>
#include <stream.h>
#include <FlexLexer.h>
#include <string.h>
#include "Vector.h"
#include "types.h"
#include "ValueType.h"
#include "PasNode.h"
#include "javacls/javacls.h"
#include "javacls/javainst.h"
#include "PascalLexer.h"
#include "Compiler.h"
#include "VarTable.h"

JavaClass cls;
JavaCode *javaCode;
ReturnNode *retVal=NULL;
ProcedureNode *aktProcedure;
JavaLabel breakLabel;
extern Compiler *compiler;
extern char *className;

void ProgramNode::generateStaticInit(){
  JavaMethod *met = cls.addMethod("<clinit>","()V",
	JavaClass::ACC_FINAL|JavaClass::ACC_STATIC);
  javaCode=new JavaCode(met);
  javaCode->setLimits(2,0);
  compiler->globalTable()->initArrays(FALSE);
  javaCode->gen(RETURN);
} 

void ProgramNode::generate(){
  cls.create(name);
  decls->generate();
  mainProc->generate();
  cls.addDefaultInitMethod();
  generateStaticInit();
  cls.save();
  cls.dump();
}
void ArrayNode::initArray(int local){
  javaCode->genIntConst(endIndex-beginIndex+1);
  switch (type&(~ATYP)){
    case ITYP:
	javaCode->genNewArray('I');
        break;
    case BTYP:
	javaCode->genNewArray('Z');
        break;
    case CTYP:
	javaCode->genNewArray('C');
        break;
    case STYP:
	javaCode->genNewArray("java/lang/String");
        break;
    default:
        compiler->genError("Unknown type of array");
  }
  if(local==TRUE){ 
    javaCode->genStore(PasNode::javaStoreType(type),(JavaUShort)localAdr);
  }
  else{
    javaCode->genField(PUTSTATIC, className, name, 
		PasNode::javaType(type));
  }
}
void ReturnNode::initArray(int local){
  if((type&ATYP)==0){
    return;
  }
  javaCode->genIntConst(endIndex-beginIndex+1);
  switch (type&(~ATYP)){
    case ITYP:
	javaCode->genNewArray('I');
        break;
    case BTYP:
	javaCode->genNewArray('Z');
        break;
    case CTYP:
	javaCode->genNewArray('C');
        break;
    case STYP:
	javaCode->genNewArray("java/lang/String");
        break;
    default:
        compiler->genError("Unknown type of array");
  }
  if(local==TRUE){ 
    javaCode->genStore(PasNode::javaStoreType(type),(JavaUShort)localAdr);
  }
  else{
    compiler->genError("Return value in global table");
  }
}
char *PasNode::javaType(int type){
  switch(type){
    case ITYP:
      return "I";
    case CTYP:
      return "C";
    case BTYP:
      return "Z";
    case VTYP:
      return "V";
    case STYP:
      return "Ljava/lang/String;";
    case ITYP|ATYP:
      return "[I";
    case CTYP|ATYP:
      return "[C";
    case BTYP|ATYP:
      return "[Z";
    case STYP|ATYP:
      return "[Ljava/lang/String;";
    default:
      compiler->genError("Unknown type");
      return "";
  }
}
char PasNode::javaStoreType(int type){
  switch(type){
    case ITYP:
      return 'I';
    case CTYP:
      return 'I';
    case BTYP:
      return 'I';
    case VTYP:
      return 'V';
    case STYP:
      return 'L';
    case ITYP|ATYP:
      return '[';
    case CTYP|ATYP:
      return '[';
    case BTYP|ATYP:
      return '[';
    case STYP|ATYP:
      return '[';
    default:
      compiler->genError("Unknown type");
      return 'I';
  }
}
    
void DeclsNode::generate(){
  int size=decls->getSize();
  int i;
  for(i=0;i<size;i++){
    TabEntry *entry = (TabEntry *)decls->elementAt(i);
    entry->generate();
  }
}

char *TabEntry::javaType(){
  return PasNode::javaType(type);
}
void TabEntry::generate(){
  compiler->genError("Calling of method TabEntry::generate");
}

void Variable::generate(){
  cls.addField(name,PasNode::javaType(type),
	JavaClass::ACC_PUBLIC|JavaClass::ACC_STATIC);
}

void ArrayNode::generate(){
  cls.addField(name,PasNode::javaType(type),
	JavaClass::ACC_PUBLIC|JavaClass::ACC_STATIC);
}

char *ProcedureNode::javaMethodType(){
  return parameters->javaParamsType();
}
char *ParametersNode::javaParamsType(){
  int size = variables->getSize();
  int i;
  char *desc;
  if(variables==0){
    desc=new char[3];
    strcpy(desc,"()");
    return desc;
  }
  desc=new char[2];
  strcpy(desc,"(");
  for(i=0;i<size;i++){
    char *typ=((TabEntry *)variables->elementAt(i))->javaType();
    char *desc2 = new char[strlen(desc)+strlen(typ)+3];
    strcpy(desc2,desc);
    desc2=strcat(desc2,typ);
    delete desc;
    desc=desc2;
  }
  desc=strcat(desc,")");
  return desc;
}
void ProcedureNode::generate(){
  aktProcedure=this;
  char *desc1=javaMethodType();
  char *desc2=PasNode::javaType(type);
  char *desc=new char[strlen(desc1)+strlen(desc2)+1];
  strcpy(desc,desc1);
  desc=strcat(desc,desc2);
  JavaMethod *met = cls.addMethod(name,desc,
	JavaClass::ACC_PUBLIC|JavaClass::ACC_STATIC);
  setDesc(desc);
  retVal=localTable->returnValue();
  javaCode=new JavaCode(met);
  int size1=code->getStackSize();
  int size2=retVal->getStackSize();
  size1=(size1>size2)?size1:size2;
  size1=(size1>2)?size1:2;
  javaCode->setLimits(size1,localTable->localsSize());
  localTable->initArrays(TRUE);
  code->generate();
  generateReturn();
  retVal=NULL;
  aktProcedure=NULL;
}
void MainProcedureNode::generate(){
  aktProcedure=this;
  JavaMethod *met = cls.addMethod(name,desc,
	JavaClass::ACC_PUBLIC|JavaClass::ACC_STATIC);
  retVal=localTable->returnValue();
  javaCode=new JavaCode(met);
  int size1=code->getStackSize();
  int size2=retVal->getStackSize();
  size1=(size1>size2)?size1:size2;
  size1=(size1>2)?size1:2;
  javaCode->setLimits(size1,localTable->localsSize());
  code->generate();
  generateReturn();
  retVal=NULL;
  aktProcedure=NULL;
}
void ProcedureNode::generateReturn(){
  if(retVal==NULL || retVal->getNodeType() == RETPTYP){
    javaCode->gen(RETURN);
  }
  else{
    int t=retVal->getType();
    javaCode->genLoad(PasNode::javaStoreType(t),
	(JavaUShort)retVal->getLocalAdr());
    if(t==ITYP||t==CTYP||t==BTYP){
      javaCode->gen(IRETURN);
    }
    else if(t==STYP || (t&ATYP)!=0){
      javaCode->gen(ARETURN);
    }
    else {
      compiler->genError("Unknown type of return value");
    }
  }
}

int BinaryNode::getStackSize(){
  int l=loperand->getStackSize();
  int p=poperand->getStackSize()+1;
  return (l>p)?l:p;
}

int ReturnNode::getStackSize(){
  return PasNode::typesLength(type);
}
int UnaryNode::getStackSize(){
  return operand->getStackSize();
}
int LetStmntNode::getStackSize(){
  return expr->getStackSize()+address->getStackSize();
}
int CallStmntNode::getStackSize(){
  return functCall->getStackSize();
}
int WhileStmntNode::getStackSize(){
  int e=expr->getStackSize();
  int s=stm->getStackSize();
  return (e>s)?e:s;
}
int IfStmntNode::getStackSize(){
  int e=expr->getStackSize();
  int s=stm->getStackSize();
  return (e>s)?e:s;
}
int IfElseStmntNode::getStackSize(){
  int e=expr->getStackSize();
  int s1=stm1->getStackSize();
  int s2=stm2->getStackSize();
  e=(e>s1)?e:s1;
  return (e>s2)?e:s2;
}
int BreakStmntNode::getStackSize(){
  return 0;
}
int ReturnStmntNode::getStackSize(){
  return 2;
}
int EmptyStmntNode::getStackSize(){
  return 0;
}
int StatementNode::getStackSize(){
  compiler->genError("Call of method StatementNode::getStackSize()");
  return 0;
}

int ArrayAdrNode::getStackSize(){
  if(index!=NULL){
    return index->getStackSize()+2;
  }
  else{
    return 0;
  }
}
int VariableAdrNode::getStackSize(){
  return 0;
}
int ReturnAdrNode::getStackSize(){
  if(index!=NULL){
    return index->getStackSize() + 2;
  }
  else {
    return 0;
  }
}
int LeftValueNode::getStackSize(){
  compiler->genError("Call of method LeftValueNode::getStackSize()");
  return 0;
}
int BoolConstNode::getStackSize(){
  return 1;
}
int IntConstNode::getStackSize(){
  return 1;
}
int StrConstNode::getStackSize(){
  return 1;
}
int CharConstNode::getStackSize(){
  return 1;
}
int CallNode::getStackSize(){
  int p = parameters->getStackSize();
  p=(p==0)?1:p;
  if(index!=NULL){
    int p2 = index->getStackSize() + 2;
    p=(p>p2)?p:p2;
  }
  return p;
}

int RealParametersNode::getStackSize(){
  int size = params->getSize();
  int p = 0;
  int i;
  for(i=0;i<size;i++){
    int p2=((ExprNode *)params->elementAt(i))->getStackSize()+i;
    p=(p>p2)?p:p2;
  }
  return p;
}
int PasNode::typesLength(int type){
  switch(type){
    case ITYP:
      return 1;
    case CTYP:
      return 1;
    case BTYP:
      return 1;
    case STYP:
      return 1;
    case VTYP:
      return 0;
    case ITYP|ATYP:
      return 1;
    case CTYP|ATYP:
      return 1;
    case STYP|ATYP:
      return 1;
    case BTYP|ATYP:
      return 1;
    default: 
      compiler->genError("Unknown type");
  }
}

int VariableValNode::getStackSize(){
  if(index==NULL){
    return PasNode::typesLength(type);
  }
  else{
    return index->getStackSize()+3;
  }
}

int ArrayValNode::getStackSize(){
  if(index!=NULL){
    return index->getStackSize()+2;
  }
  else{
    return PasNode::typesLength(type);
  }
}

int SequenceNode::getStackSize(){
  int size=stms->getSize();
  int i;
  int p=0;
  int p2=0;
  for(i=0;i<size;i++){
    p2=((StatementNode *)stms->elementAt(i))->getStackSize();
    p=(p>p2)?p:p2;
  }
  return p;
}

void ExprNode::generate(){
  compiler->genError("Call of method ExprNode::generate()");
}
void UnaryNode::generate(){
  operand->generate();
  switch(oper){
    case MINUS:
               javaCode->gen(INEG);
               break;
    case NOT:
               JavaLabel lt,le;
               le=javaCode->getLabel();
               lt=javaCode->getLabel();
               javaCode->genJump(IFEQ,lt);
               javaCode->gen(ICONST_0);
               javaCode->genJump(GOTO,le);
               javaCode->genLabel(lt);
               javaCode->gen(ICONST_1);
               javaCode->genLabel(le);
               break;
    default:
      compiler->genError("Unknown unary operator");
  }
}
               
		
void RightValueNode::generate(){
  compiler->genError("Call of method RightValueNode::generate");
}
void LetStmntNode::generate(){
  address->generateBegin();
  expr->generate();
  address->generate();
}
void WhileStmntNode::generate(){
  JavaLabel lpodm,lbegin,old;
  old=breakLabel;

  lpodm=javaCode->getLabel();
  lbegin=javaCode->getLabel();
  breakLabel=javaCode->getLabel();

  javaCode->genJump(GOTO,lpodm);
  javaCode->genLabel(lbegin);
  stm->generate();
  javaCode->genLabel(lpodm);  
  expr->generate();
  javaCode->genJump(IFNE,lbegin);
  javaCode->genLabel(breakLabel);  

  breakLabel=old;
}
void IfStmntNode::generate(){
  expr->generate();
  JavaLabel lt;
  lt=javaCode->getLabel();
  javaCode->genJump(IFEQ,lt);
  stm->generate();
  javaCode->genLabel(lt);
}
void StatementNode::generate(){
}
void LeftValueNode::generate(){
  compiler->genError("Call of method LeftValueNode::generate");
}
void VariableValNode::generate(){
  int typ=type;
  if(index!=NULL)
    type=STYP;
  if(local==TRUE){ 
    javaCode->genLoad(PasNode::javaStoreType(type),
	(JavaUShort)val->getLocalAdr());
  }
  else{
    javaCode->genField(GETSTATIC, className, val->getName(), 
		PasNode::javaType(type));
  }
  type=typ;
  if(index!=NULL){
    index->generate();
    javaCode->genCall(INVOKESTATIC,"Methods","charAt",
		"(Ljava/lang/String;I)C");
  }
}

int ArrayValNode::loadInstr(int type){
  switch(type&(~ATYP)){
    case ITYP:
      return IALOAD;
    case BTYP:
      return BALOAD;
    case CTYP:
      return CALOAD;
    case STYP:
      return AALOAD;
    default:
      compiler->genError("Unknown type of array");
      return IALOAD;
  }
}
int ArrayAdrNode::storeInstr(int type){
  switch(type&(~ATYP)){
    case ITYP:
      return IASTORE;
    case BTYP:
      return BASTORE;
    case CTYP:
      return CASTORE;
    case STYP:
      return AASTORE;
    default:
      compiler->genError("Unknown type of array");
      return IASTORE;
  }
}
void ArrayValNode::generate(){
  if(local==TRUE){ 
    javaCode->genLoad('[',(JavaUShort)val->getLocalAdr());
  }
  else{
    javaCode->genField(GETSTATIC, className, val->getName(), 
		PasNode::javaType(type|ATYP));
  }
  if(index!=NULL){
    index->generate();
    if(val->getBeginIndex()!=0){
      javaCode->genIntConst(val->getBeginIndex());
      javaCode->gen(ISUB);
    }
    javaCode->gen((JavaInstr)ArrayValNode::loadInstr(type));
  }
}

void CallNode::generate(){
  parameters->generate();
  procedure->generateCall(parameters);
  if(index!=NULL){
    index->generate();
    if(procedure->getBeginIndex()!=0){
      javaCode->genIntConst(procedure->getBeginIndex());
      javaCode->gen(ISUB);
    }
    javaCode->gen((JavaInstr)ArrayValNode::loadInstr(type));
  }
}

void ProcedureNode::generateCall(RealParametersNode *param){
  javaCode->genCall(INVOKESTATIC,className,name,desc);
}
void RealParametersNode::generate(){
  int size=params->getSize();
  int i;
  for(i=0;i<size;i++){
    ((ExprNode *)params->elementAt(i))->generate();
  }
}

void VariableAdrNode::generate(){
  if(local==TRUE){ 
    javaCode->genStore(PasNode::javaStoreType(type),
	(JavaUShort)val->getLocalAdr());
  }
  else{
    javaCode->genField(PUTSTATIC, className, val->getName(), 
		PasNode::javaType(type));
  }
}
void ArrayAdrNode::generateBegin(){
  if(index!=NULL){
    if(local==TRUE){ 
      javaCode->genLoad('[',(JavaUShort)val->getLocalAdr());
    }
    else{
      javaCode->genField(GETSTATIC, className, val->getName(), 
  		PasNode::javaType(type|ATYP));
    }
    index->generate();
    if(val->getBeginIndex()!=0){
      javaCode->genIntConst(val->getBeginIndex());
      javaCode->gen(ISUB);
    }
  }
}
void ReturnAdrNode::generateBegin(){
  if(index!=NULL){
    if(local==TRUE){ 
      javaCode->genLoad('[',(JavaUShort)val->getLocalAdr());
    }
    else{
      compiler->genError("Global return value");
    }
    index->generate();
    if(val->getBeginIndex()!=0){
      javaCode->genIntConst(val->getBeginIndex());
      javaCode->gen(ISUB);
    }
  }
}

void ReturnAdrNode::generate(){
  if(index==NULL){
    if(local==TRUE){ 
      javaCode->genStore(PasNode::javaStoreType(type),
  	(JavaUShort)val->getLocalAdr());
    }
    else{
      compiler->genError("Global return value");
    }
  }
  else{
    javaCode->gen((JavaInstr)ArrayAdrNode::storeInstr(type));
  }
}

void ArrayAdrNode::generate(){
  if(index==NULL){
    if(local==TRUE){ 
      javaCode->genStore(PasNode::javaStoreType(type),
  	(JavaUShort)val->getLocalAdr());
    }
    else{
      javaCode->genField(PUTSTATIC, className, val->getName(), 
		PasNode::javaType(type));
    }
  }
  else{
    javaCode->gen((JavaInstr)ArrayAdrNode::storeInstr(type));
  }
}


void IfElseStmntNode::generate(){
  expr->generate();
  JavaLabel lt,le;
  le=javaCode->getLabel();
  lt=javaCode->getLabel();
  javaCode->genJump(IFEQ,lt);
  stm1->generate();
  javaCode->genJump(GOTO,le);
  javaCode->genLabel(lt);
  stm2->generate();
  javaCode->genLabel(le);
}
void BreakStmntNode::generate(){
  javaCode->genJump(GOTO,breakLabel);
}
void ReturnStmntNode::generate(){
  if(aktProcedure!=NULL){
    aktProcedure->generateReturn();
  }
}
void CallStmntNode::generate(){
  functCall->generate();
  if(functCall->getType()!=VTYP){
    javaCode->gen(POP);
  }
}
void EmptyStmntNode::generate(){
}
void CharConstNode::generate(){
  javaCode->genIntConst((JavaInt) charVal);
}
void IntConstNode::generate(){
  javaCode->genIntConst((JavaInt) intVal);
}
void StrConstNode::generate(){
  javaCode->genStringConst(strVal);
}
void BoolConstNode::generate(){
  if(boolVal==1){
    javaCode->gen(ICONST_1);
  }
  else{
    javaCode->gen(ICONST_0);
  }
}
void SequenceNode::generate(){
  int size=stms->getSize();
  int i;
  for(i=0;i<size;i++){
    ((StatementNode *)stms->elementAt(i))->generate();
  }
}
void BinaryNode::generateRel(int i){
  JavaLabel lt,le;
  le=javaCode->getLabel();
  lt=javaCode->getLabel();
  javaCode->genJump((JavaInstr)i,lt);
  javaCode->gen(ICONST_0);
  javaCode->genJump(GOTO,le);
  javaCode->genLabel(lt);
  javaCode->gen(ICONST_1);
  javaCode->genLabel(le);
}
void BinaryNode::generate(){
  loperand->generate();
  poperand->generate();
  switch(oper){
    case PLUS:     javaCode->gen(IADD);
                   break;
    case MINUS:    javaCode->gen(ISUB);
                   break;
    case OR:       javaCode->gen(IOR);
                   break;
    case KRAT:     javaCode->gen(IMUL);
                   break;
    case DELENO:   javaCode->gen(IDIV);
                   break;
    case DIV:      javaCode->gen(IDIV);
                   break;
    case MOD:      javaCode->gen(IREM);
                   break;
    case AND:      javaCode->gen(IAND);
                   break;
    case LESS:     
                   generateRel(IF_ICMPLT);
                   break;
    case LESSEQ:   
                   generateRel(IF_ICMPLE);
                   break;
    case EQ:       
                   generateRel(IF_ICMPEQ);
                   break;
    case GREATEREQ:
                   generateRel(IF_ICMPGE);
                   break;
    case GREATER:  
                   generateRel(IF_ICMPGT);
                   break;
    case NOEQ:     
                   generateRel(IF_ICMPNE);
                   break;
    default:
      compiler->genError("Unknown binary operator");
  }
}

