#include <stdlib.h>
#include <stream.h>
#include <FlexLexer.h>

#include "Vector.h"
#include "types.h"
#include "ValueType.h"
#include "PasNode.h"
#include "VarTable.h"
#include "PascalLexer.h"
#include "Compiler.h"

extern Compiler *compiler;


int PasNode::compatible(int typ1, int typ2){
  if((typ1 & ETYP) != 0 || (typ2 & ETYP) != 0){ 
    return ERROR;
  }
  if(typ1==typ2){
    return TRUE;
  }
  if(( (typ1&CTYP)!=0 && (typ2&ITYP)!=0 || (typ1&ITYP)!=0 && (typ2&CTYP)!=0 )
	&& ((typ1&ATYP)==(typ2&ATYP)) ){
    return TRUE;
  }
  return FALSE;
}


int ProcedureNode::compatible(RealParametersNode *real){
  return parameters->compatible(real);
}


int ParametersNode::compatible(RealParametersNode *real){
  Vector *r=real->getParams();
  int s=variables->getSize();
  int i;
  int t1,t2;
  if(s != r->getSize()){
    return FALSE;
  }
  for(i=0;i<s;i++){
    t1=((TabEntry *)(variables->elementAt(i)))->getType();
    t2=((TabEntry *)(r->elementAt(i)))->getType();
    if(PasNode::compatible(t1, t2)==FALSE){
      return FALSE;
    }
  }
  return TRUE;
}


CallNode* CallNode::construct(char *name, RealParametersNode *param){
  if(name==NULL){
    return new CallNode(ETYP);
  }
  TabEntry *funct = compiler->globalTable()->getVal(name);
  if((funct->getType() & ETYP)!=0){
    compiler->yyerrorShort();
    cerr<<"Function or procedure "<<name<<" not declared"<<endl;
    return new CallNode(ETYP);
  }
  if(funct->getNodeType() != FUNCTYP && funct->getNodeType() != PROCTYP){
    compiler->yyerrorShort();
    cerr<<"Identifier "<<name<<" not defined as function"<<endl;
    return new CallNode(ETYP);
  }
  return ((ProcedureNode *)funct)->construct(param);
}


CallNode* MainProcedureNode::construct(RealParametersNode *param){
  compiler->yyerrorShort();
  cerr<<"Can not call procedure main"<<endl;
  return new CallNode(ETYP);
}

CallNode* ProcedureNode::construct(RealParametersNode *param){
  if(compatible(param) == TRUE){
    return new CallNode(this, param,NULL);
  }
  else{
    compiler->yyerrorShort();
    cerr<<"Uncompatible parameters to function or procedure "<<
	name<<" call"<<endl;
    return new CallNode(ETYP);
  }
}

BinaryNode::BinaryNode(int o, ExprNode *l, ExprNode *p, int t){
  oper = o;
  loperand = l;
  poperand = p;
  type=t;
}

BinaryNode::BinaryNode(int o, ExprNode *l, ExprNode *p){
  oper = o;
  loperand = l;
  poperand = p;
  if(l!=NULL &&p!=NULL){
    type=l->getType();
  }
  else{
    type=ETYP;
  }
}


TabEntry *TabEntry::searchEntry(char *name, int *local){
  TabEntry *entry;
  int typ, nodeTyp;
  if(compiler->localTable()!=NULL){
    entry = compiler->localTable()->getVal(name);
    typ = entry->getType();
    nodeTyp=entry->getNodeType();
    *local=TRUE;
  }
  else{
    typ = ETYP;
    nodeTyp=ERRORTYP;
    *local=FALSE;
  }
  if((typ & ETYP)!=0 || nodeTyp==ERRORTYP){
    entry = compiler->globalTable()->getVal(name);
    *local=FALSE;
  }
  return entry;
}

 
RightValueNode* TabEntry::constructRightValue(char *name,ExprNode *expr,RealParametersNode *realPar){
  int nodeType;
  TabEntry *entry;
  int local;
  if(name==NULL){
    return new RightValueNode(ETYP);
  }
  entry=TabEntry::searchEntry(name,&local);
  return entry->getRightValue(expr,realPar,local);
}



RightValueNode *TabEntry::getRightValue(ExprNode *expr,
		RealParametersNode *realPar,int local){
  compiler->yyerrorShort();
  cerr<<"Function or variable "<<name<<" not declared"<<endl;
  return new RightValueNode(ETYP);
}

RightValueNode *Variable::getRightValue(ExprNode *expr,
		RealParametersNode *realPar,int local){
  if(expr!=NULL){
    if(type==STYP){
      return new VariableValNode(this, CTYP, expr, local);
    }
    else{
      compiler->yyerrorShort();
      cerr<<"Can not index simple type variable "<<name<<endl;
          return new RightValueNode(ETYP);
    }
  }
  else {
    return new VariableValNode(this, type, NULL, local);
  }
}

RightValueNode *ProcedureNode::getRightValue(ExprNode *expr,
                RealParametersNode *realPar,int local){
  if(nodeType==PROCTYP){
    compiler->yyerrorShort();
    cerr<<"Procedure "<<name<<" does not return value"<<endl;
    return new RightValueNode(ETYP);
  }
  else if(nodeType==FUNCTYP){
    if(compatible(realPar) == TRUE){
      if(expr!=NULL){
        if((type&ATYP)!=0){
          return new CallNode(this, type&(~ATYP),realPar, expr);
        }
        else{
          compiler->yyerrorShort();
          cerr<<"Can not index function "<<name<<endl;
          return new RightValueNode(ETYP);
        }
      }
      else {
        return new CallNode(this, type, realPar, NULL);
      }
    }
    else{
      compiler->yyerrorShort();
      cerr<<"Uncompatible parameters to function or procedure "<<
          name<<" call"<<endl;
      return new RightValueNode(ETYP);
    }
  }
  else {
    compiler->yyerrorShort();
    cerr<<name<<"is not function"<<endl;
    return new RightValueNode(ETYP);
  }
}

RightValueNode *ProgramNode::getRightValue(ExprNode *expr,
		RealParametersNode *realPar,int local){
    compiler->yyerrorShort();
    cerr<<"Program "<<name<<" does not return value"<<endl;
    return new RightValueNode(ETYP);
}

RightValueNode *ReturnNode::getRightValue(ExprNode *expr,
		RealParametersNode *realPar,int local){
 
    TabEntry *entry = compiler->globalTable()->getVal(name);
    if(entry->getNodeType()==PROCTYP || entry->getNodeType()==FUNCTYP){
      return entry->getRightValue(expr,realPar,local);
    }
    compiler->yyerrorShort();
    cerr<<"Can not use return value"<<endl;
    return new RightValueNode(ETYP);
}

RightValueNode *ArrayNode::getRightValue(ExprNode *expr,
		RealParametersNode *realPar,int local){
  if(expr==NULL){
    return new ArrayValNode(this,type,NULL,local);
  }
  else{
    return new ArrayValNode(this,type&(~ATYP),expr,local);
  }
}


LeftValueNode *TabEntry::constructLeftValue(char *name, ExprNode *expr){
  int nodeType;
  TabEntry *entry;
  int local;
  if(name==NULL){
    return new LeftValueNode(ERRORTYP,ETYP);
  }
  entry=TabEntry::searchEntry(name,&local);
  return entry->getLeftValue(expr,local);
}

LeftValueNode *TabEntry::getLeftValue(ExprNode *expr,int local){
  compiler->yyerrorShort();
  cerr<<"Variable "<<name<<" not exists"<<endl;
  return new LeftValueNode(ERRORTYP,ETYP);
}
LeftValueNode *ProgramNode::getLeftValue(ExprNode *expr,int local){
  compiler->yyerrorShort();
  cerr<<"Program "<<name<<" can not assign a value"<<endl;
  return new LeftValueNode(ERRORTYP,ETYP);
}
LeftValueNode *ProcedureNode::getLeftValue(ExprNode *expr,int local){
  compiler->yyerrorShort();
  cerr<<"Procedure or function "<<name<<" can not assign a value"<<endl;
  return new LeftValueNode(ERRORTYP,ETYP);
}

LeftValueNode *Variable::getLeftValue(ExprNode *expr,int local){
  if(expr!=NULL){
    compiler->yyerrorShort();
    cerr<<"Indexation of simple type variable "<<name<<endl;
    return new LeftValueNode(ERRORTYP,ETYP);
  }
  return new VariableAdrNode(this,type,local);
}

LeftValueNode* ArrayNode::getLeftValue(ExprNode *expr,int local){
  if(expr!=NULL){
    return new ArrayAdrNode(this,type&(~ATYP),expr,local);
  }
  else {
    return new ArrayAdrNode(this,type,expr,local);
  }
}

LeftValueNode *ReturnNode::getLeftValue(ExprNode *expr,int local){
  if(type==VTYP){
    compiler->yyerrorShort();
    cerr<<"Procedure "<<name<<" can not return value"<<endl;
    return new LeftValueNode(ERRORTYP,ETYP);
  }
  if(expr!=NULL){
    if((type&ATYP)!=0){
      return new ReturnAdrNode(this,type&(~ATYP),expr);
    }
    else{
      compiler->yyerrorShort();
      cerr<<"Procedure "<<name<<" returns simple type value"<<endl;
      return new LeftValueNode(ERRORTYP,ETYP);
    }
  }
  else{
     return new ReturnAdrNode(this,type,expr);
  }
}


UnaryNode *UnaryNode::construct(int o, ExprNode *p){
  int t;
  if(p==NULL){
   return new UnaryNode(ETYP);
  }
  t= p->getType();
  if((t&ETYP)!=0){
   return new UnaryNode(ETYP);
  }
  if(t==VTYP){
    compiler->yyerrorShort();
    cerr<<"Procedure returns no value"<<endl;
    return new UnaryNode(ETYP);
  }
  if(o==NOT){
    if(t==BTYP){
      return new UnaryNode(o, p, BTYP);
    }
    else{
      compiler->yyerrorShort();
      cerr<<"Can negate only boolean value"<<endl;
      return new UnaryNode(ETYP);
    }
  }
  if(o==MINUS){
    if(t==ITYP||t==CTYP){
      return new UnaryNode(o, p, ITYP);
    }
    else{
      compiler->yyerrorShort();
      cerr<<"Negative can be only integer values"<<endl;
      return new UnaryNode(ETYP);
    }
  }
  compiler->yyerrorShort();
  cerr<<"Not supported unary operator"<<endl;
  return new UnaryNode(ETYP);
}

  
        
BinaryNode *BinaryNode::construct(int o, ExprNode *p1, ExprNode *p2){
  int t1,t2;
  if(p1==NULL||p2==NULL){
   return new BinaryNode(ETYP);
  }
  t1= p1->getType();
  t2= p2->getType();
  if((t1&ETYP)!=0 || (t2&ETYP)!=0){
   return new BinaryNode(ETYP);
  }
  if(t1==VTYP||t2==VTYP){
    compiler->yyerrorShort();
    cerr<<"Procedure returns no value"<<endl;
    return new BinaryNode(ETYP);
  }
  if(o==PLUS || o==MINUS || o==KRAT || o==DELENO || o==DIV || o==MOD){
    if((t1==ITYP||t1==CTYP) && (t2==ITYP||t2==CTYP)){
      return new BinaryNode(o, p1, p2, ITYP);
    }
    else{
      compiler->yyerrorShort();
      cerr<<"Multiplication and addition operators can be ";
      cerr<<"used only on integer value"<<endl;
      return new BinaryNode(ETYP);
    }
  }
  if(o==LESS || o==LESSEQ || o==EQ || o==GREATEREQ || o==GREATER || o==NOEQ){
    if((t1==ITYP||t1==CTYP) && (t2==ITYP||t2==CTYP)){
      return new BinaryNode(o, p1, p2, BTYP);
    }
    else{
      compiler->yyerrorShort();
      cerr<<"Multiplication, addition and relation operators"<<endl;
      cerr<<"\tcan be used only on integer value"<<endl;
      return new BinaryNode(ETYP);
    }
  }
  if(o==AND || o==OR){
    if(t1==BTYP && t2==BTYP){
      return new BinaryNode(o, p1, p2, BTYP);
    }
    else{
      compiler->yyerrorShort();
      cerr<<"Logical operator can be used only on boolean values"<<endl;
      return new BinaryNode(ETYP);
    }
  }
  compiler->yyerrorShort();
  cerr<<"Not supported unary operator"<<endl;
  return new BinaryNode(ETYP);
}



ProcedureNode *ProcedureNode::construct(char *name, VarTable *table, 
		int function, int line){
  TabEntry *entry;
  ProcedureNode *proc;
  int typ;
  entry = compiler->globalTable()->getVal(name);
  typ = entry->getType();
  if(typ==ETYP){
    proc = new ProcedureNode(((function==TRUE)?FUNCTYP:PROCTYP),name);
    table->reset();
    proc->setTable(table);
    table->addVar(new ReturnNode(proc, ((function==TRUE)?RETTYP:RETPTYP),
		VTYP,name));
    proc->setLine(line);
    return proc;
  }
  else {
    compiler->yyerrorShort();
    cerr<<"Identifier "<<name<<" is alraedy defined"<<endl;
    cerr<<"First definition at line "<<entry->getLine()<<endl;
    return new ProcedureNode(ERRORTYP, ETYP);
  }
}



LetStmntNode *LetStmntNode::construct(LeftValueNode *adr, ExprNode *e){
  if(adr==NULL || e==NULL){
    return NULL;
  }
  if(adr->getType()==ETYP || e->getType()==ETYP){
    return new LetStmntNode();
  }
  if(e->getType()==VTYP){
    compiler->yyerrorShort();
    cerr<<"Procedure returns no value"<<endl;
    return new LetStmntNode();
  }
   
  if(PasNode::compatible(adr->getType(),e->getType())==TRUE){
    return new LetStmntNode(adr, e);
  }
  else {
    compiler->yyerrorShort();
    cerr<<"Uncompatibile values in assignment"<<endl;
    return new LetStmntNode();
  }
}



WhileStmntNode *WhileStmntNode::construct(ExprNode *e,StatementNode *s){
  if(e->getType()==VTYP){
    compiler->yyerrorShort();
    cerr<<"Procedure returns no value"<<endl;
    return new WhileStmntNode();
  }
  if(e->getType()==BTYP){
    return new WhileStmntNode(e,s);
  }
  else{
    compiler->yyerrorShort();
    cerr<<"Condition is not boolean value"<<endl;
    return new WhileStmntNode();
  }
} 



IfStmntNode *IfStmntNode::construct(ExprNode *e,StatementNode *s){
  if(e->getType()==VTYP){
    compiler->yyerrorShort();
    cerr<<"Procedure returns no value"<<endl;
    return new IfStmntNode();
  }
  if(e->getType()==BTYP){
    return new IfStmntNode(e,s);
  }
  else{
    compiler->yyerrorShort();
    cerr<<"Condition is not boolean value"<<endl;
    return new IfStmntNode();
  }
} 



IfElseStmntNode *IfElseStmntNode::construct(ExprNode *e,StatementNode *s1,
		StatementNode *s2){
  if(e->getType()==VTYP){
    compiler->yyerrorShort();
    cerr<<"Procedure returns no value"<<endl;
    return new IfElseStmntNode();
  }
  if(e->getType()==BTYP){
    return new IfElseStmntNode(e,s1,s2);
  }
  else{
    compiler->yyerrorShort();
    cerr<<"Condition is not boolean value"<<endl;
    return new IfElseStmntNode();
  }
} 


Vector *TabEntry::makeEntries(Vector *v, ValueType *t){
  int s=v->getSize();
  int i;
  if((t->typ&ATYP)!=0){
    for(i=0;i<s;i++){
      ArrayNode *ar=new ArrayNode(t->typ,((NameLine *)v->elementAt(i))->name,
		t->begin,t->end);
      ar->setLine(((NameLine *)v->elementAt(i))->line);
      v->setElementAt(i,ar);
    }
  }
  else {
    for(i=0;i<s;i++){
      Variable *var=new Variable(t->typ,
		((NameLine *)v->elementAt(i))->name,
		((NameLine *)v->elementAt(i))->line);
      v->setElementAt(i,var);
    }
  } 
  return v;
}


TabEntry::TabEntry(){
  type=ETYP;
  nodeType=ERRORTYP;
  name=NULL;
  used=FALSE;
  line = 0;
  isParameter=FALSE;
}

TabEntry::TabEntry(int nt, int t){
  type=t;
  nodeType=nt;
  name=NULL;
  used=FALSE;
  line = 0;
  isParameter=FALSE;
}

TabEntry::TabEntry(char *n){
  type=VTYP;
  nodeType=ERRORTYP;
  name=n;
  used=FALSE;
  line = 0;
  isParameter=FALSE;
}

TabEntry::TabEntry(int nt,char *n){
  nodeType=nt;
  type=ETYP;
  name=n;
  used=FALSE;
  line = 0;
  isParameter=FALSE;
}

TabEntry::TabEntry(int nt,int t, char *n){
  nodeType=nt;
  type=t;
  name=n;
  used=FALSE;
  line = 0;
  isParameter=FALSE;
}

TabEntry::TabEntry(int nt, int t,char *n, int l){
  nodeType=nt;
  type=t;
  name=n;
  used=FALSE;
  line = l;
  isParameter=FALSE;
}

ProgramNode *ProgramNode::construct(NameLine *nl){
  if(nl->name==NULL){
    return new ProgramNode(nl->line,nl->name);
  }
  if(strcmp(nl->name, compiler->getLexer()->getFilenameShort())!=0){
    compiler->yyerrorShort();
    cerr<<"Name of program "<<nl->name;
    cerr<<" is different from name of file ";
    cerr<<compiler->getLexer()->getFilenameShort()<<endl;
  }
  return new ProgramNode(nl->line,nl->name);
}


BreakStmntNode *BreakStmntNode::construct(int isWhile){
  if(isWhile==FALSE){
    compiler->yyerrorShort();
    cerr<<"Break not in block in while"<<endl;
  }
  else {
    return new BreakStmntNode();
  }
}


ProgramNode::ProgramNode(char *n):TabEntry(PROGTYP,VTYP,n){
  mainProc = new MainProcedureNode();
}

ProgramNode::ProgramNode(int l, char *n):TabEntry(PROGTYP,VTYP,n,l){
  mainProc = new MainProcedureNode();
}


MainProcedureNode *ProgramNode::getMainProc(){
  return mainProc;
}


void ProgramNode::setCode(SequenceNode *c){
  mainProc->setCode(c);
  code = c;
}


void MainProcedureNode::setParameters(ParametersNode *par){
  compiler->yyerrorShort();
  cerr<<"Function main has constant parameters"<<endl;
}

void MainProcedureNode::setTable(VarTable *l){
  compiler->yyerrorShort();
  cerr<<"Function main has not local variables"<<endl;
}

void MainProcedureNode::setLocalDecls(DeclsNode *d){
  compiler->yyerrorShort();
  cerr<<"Function main has not local variables"<<endl;
}

void MainProcedureNode::setDesc(char *d){
  compiler->yyerrorShort();
  cerr<<"Function main has constant descriptor"<<endl;
}

MainProcedureNode::MainProcedureNode():ProcedureNode(PROCTYP,mainName){
  desc=mainDesc;
  localTable=new VarTable();
  localTable->addVar(new ReturnNode(this, RETPTYP,VTYP,name));
  Vector *v=new Vector();
  ArrayNode *arguments=new ArrayNode(STYP|ATYP,argName,0);
  v->add(arguments);
  localTable->addVar(arguments);
  parameters=new ParametersNode(v);
}

char *MainProcedureNode::mainName = "main";
char *MainProcedureNode::mainDesc = "([Ljava/lang/String;)V";
char *MainProcedureNode::argName = "argv";

MainProcedureNode::MainProcedureNode(SequenceNode *seq):ProcedureNode(PROCTYP, 
		mainName){
  localTable=new VarTable();
  localTable->addVar(new ReturnNode(this, RETPTYP,VTYP,name));
  Vector *v=new Vector();
  ArrayNode *arguments=new ArrayNode(STYP|ATYP,argName,0);
  v->add(arguments);
  localTable->addVar(arguments);
  parameters=new ParametersNode(v);
  code=seq;
}

