struct NameLine{
  char *name;
  int line;
  NameLine(char *n, int l){
   name=n;
   line=l;
  }
};

class VarTable;

class PasNode{
  public:
    static int compatible(int typ1, int typ2);
    virtual void generate(){
    }
    static char *javaType(int type);
    static char javaStoreType(int type);
    static int typesLength(int type);
};

class LeftValueNode: public PasNode{
  private:
    int nodeType;
  protected:
    int type;
    int local;
  public:
    LeftValueNode(){
      type = ETYP;
      nodeType = ERRORTYP;
      local=FALSE;
    }
    LeftValueNode(int nt, int t){
      type = t;
      nodeType = nt;
      local=FALSE;
    }
    void setLocal(int l){
      local=l;
    }
    int getLocal(){
      return local;
    }
    int getType(){
      return type;
    }
   
    int getNodeType(){
      return nodeType;
    }
    int setType(int t){
      type = t;
    }
    int setNodeType(int nt){
      nodeType = nt;
    }
    virtual int getStackSize();
    virtual void generate();
    virtual void generateBegin(){
    }
};

class RealParametersNode;
class ExprNode;
class VariableAdrNode;
class ReturnAdrNode;
class ArrayAdrNode;
class VariableValNode;
class Variable;
class ReturnNode;
class CallNode;
class MainProcedureNode;

class ExprNode: public PasNode{
  protected:
    int type;
  public:
    ExprNode(){
      type=ETYP;
    }
    ExprNode(int t){
      type = t;
    }
    int getType(){
      return type;
    }
    virtual int getStackSize(){
    };
    virtual void generate();
};
class StatementNode: public PasNode{
  public:
    virtual int getStackSize();
    virtual void generate();
};

class SequenceNode: public PasNode{
private: 
  Vector *stms;
public:
  SequenceNode(Vector *s){
    stms = s;
  }
  virtual int getStackSize();
  virtual void generate();
};

class RightValueNode: public ExprNode{
  protected:
    int local;
  public:
    RightValueNode(int t):ExprNode(t){
      local=FALSE;
    }
    RightValueNode():ExprNode(ETYP){
      local=FALSE;
    }
    void setLocal(int l){
      local=l;
    }
    int getLocal(){
      return local;
    }
    virtual void generate();
};

class TabEntry: public PasNode{
  protected:
    int type;
    int nodeType;
    char * name;
    int used;
    int line;
    static TabEntry *searchEntry(char *name, int *local);
    int localAdr;
    int isParameter;
  public:
    TabEntry();
    TabEntry(int nt, int t);
    TabEntry(char *n);
    TabEntry(int nt,char *n);
    TabEntry(int nt,int t, char *n);
    TabEntry(int nt, int t,char *n, int l);
    int getType(){
      return type;
    }
    void setType(int i){
      type=i;
    }
    void setLocalAdr(int i){
      localAdr=i;
    }
    void setParam(int p){
      isParameter=p;
    }
    int isParam(){
      return isParameter;
    }
    int getNodeType(){
      return nodeType;
    }
    void setNodeTyp(int i){
      nodeType=i;
    }
    void setName(char *n){
      name=n;
    }
    char *getName(){
      return name;
    }
    void setUsed(){
      used=TRUE;
    }
    void setNotUsed(){
      used=FALSE;
    }
    int getUsed(){
      return used;
    }
    int getLocalAdr(){
      return localAdr;
    }
    int getLine(){
      return line;
    }
    void setLine(int l){
      line=l;
    }
    static Vector *makeEntries(Vector *v, ValueType *t);
    static LeftValueNode *constructLeftValue(char *name, ExprNode *expr);
    static RightValueNode *constructRightValue(char *name,ExprNode *expr,
		RealParametersNode *realPar);
    virtual LeftValueNode *getLeftValue(ExprNode *expr, int local);
    virtual RightValueNode *getRightValue(ExprNode *expr,
		RealParametersNode *realPar, int local);
    virtual void generate();
    char *javaType();
    virtual void initArray(int local){
    }
};

class DeclsNode: public PasNode{
  private:
    Vector *decls;
  public:
    virtual void generate();
    DeclsNode(Vector *d){
      decls=d;
    }
};

class ProgramNode: public TabEntry{
  DeclsNode *decls;
  SequenceNode *code;
  MainProcedureNode *mainProc;
  public:
    ProgramNode(char *n);
    ProgramNode(int l, char *n);
    MainProcedureNode *getMainProc();
    void setCode(SequenceNode *c);
    void setDeclarations(DeclsNode *d){
      decls=d;
    }
    static ProgramNode *construct(NameLine *nl);
    virtual void generate();
    void generateStaticInit();
    virtual RightValueNode *getRightValue(ExprNode *expr,
		RealParametersNode *realPar,int local);
    virtual LeftValueNode *getLeftValue(ExprNode *expr,int local);
};

class VariableValNode: public RightValueNode{
  private:
    Variable *val;
    ExprNode *index;
  public:
    VariableValNode(Variable *v, int t, ExprNode *i, int l):RightValueNode(t){
      val=v;
      local=l;
      index=i;
    }
    virtual int getStackSize();
    virtual void generate();
};

class VariableAdrNode: public LeftValueNode{
  private:
    Variable *val;
  public:
    VariableAdrNode(Variable *v, int t, int l):LeftValueNode(VARTYP,t){
      val=v;
      local=l;
    }
    virtual int getStackSize();
    virtual void generate();
};

class ReturnAdrNode: public LeftValueNode{
  private:
    ReturnNode *val;
    ExprNode *index;
  public:
    ReturnAdrNode(ReturnNode *v, int t,ExprNode *i):LeftValueNode(RETTYP,t){
      val=v;
      index=i;
      local=TRUE;
    }
    virtual int getStackSize();
    virtual void generate();
    virtual void generateBegin();
};

class Variable: public TabEntry{
  public:
    Variable(char *n):TabEntry(VARTYP,ETYP,n){
    }
    Variable(int t,char *n):TabEntry(VARTYP,t,n){
    }
    Variable(int t,char *n, int l):TabEntry(VARTYP,t,n,l){
    }
    virtual RightValueNode *getRightValue(ExprNode *expr,
		RealParametersNode *realPar,int local);
    virtual LeftValueNode *getLeftValue(ExprNode *expr,int local);
    virtual void generate();
};

class ParametersNode: public PasNode{
  private:
    Vector *variables;
  public:
    ParametersNode(Vector *v){
      variables=v;
    }
    int compatible(RealParametersNode *real);
    char *javaParamsType();
};

class ProcedureNode: public TabEntry{
  protected:
    VarTable *localTable;
    ParametersNode *parameters;
    SequenceNode *code;
    DeclsNode *localDecls;
    char *desc;
    int beginIndex;
    int endIndex;
  public:
    ProcedureNode(int nt, int t):TabEntry(nt, t){
    }
    ProcedureNode(int nt, char *n):TabEntry(nt, VTYP,n){
    }
    ProcedureNode(int nt, int t,char *n,ParametersNode *par, 
	SequenceNode *seq):TabEntry(nt,t,n){
      code=seq;
      parameters=par;
    }
    void setCode(SequenceNode *seq){
      code=seq;
    }
    virtual void setParameters(ParametersNode *par){
      parameters=par;
    }
    int getBeginIndex(){
      return beginIndex;
    }
    int getEndIndex(){
      return endIndex;
    }
    void setType(ValueType *t){
      beginIndex=t->begin;
      endIndex=t->end;
      type=t->typ;
    }
    virtual void setTable(VarTable *l){
      localTable=l;
    }
    VarTable *getLocalTable(){
      return localTable;
    }
    virtual void setLocalDecls(DeclsNode *d){
      localDecls=d;
    }
    virtual void setDesc(char *d){
      desc=d;
    }
    char *getDesc(){
      return desc;
    }
    static ProcedureNode *construct(char *name, VarTable *table, 
		int function, int line);

    virtual RightValueNode *getRightValue(ExprNode *expr,
		RealParametersNode *realPar,int local);
    virtual LeftValueNode *getLeftValue(ExprNode *expr,int local);
    char *javaMethodType();
    virtual CallNode *construct(RealParametersNode *param);
    virtual int compatible(RealParametersNode *real);
    virtual void generateReturn();
    virtual void generate();
    virtual void generateCall(RealParametersNode *param);
};


class MainProcedureNode: public ProcedureNode{
  private:
  public:
    static char *mainName;
    static char *mainDesc;
    static char *argName;

    MainProcedureNode();
    MainProcedureNode(SequenceNode *seq);
    virtual void setParameters(ParametersNode *par);
    virtual void setTable(VarTable *l);
    virtual void setLocalDecls(DeclsNode *d);
    virtual CallNode *construct(RealParametersNode *param);
    void setDesc(char *d);
    char *getDesc(){
      return desc;
    }
    virtual void generate();
};

class ReturnNode: public TabEntry{
  private:
    int beginIndex;
    int endIndex;
    ProcedureNode *funct;
  public:
    ReturnNode(ProcedureNode *f, int nt, char *n):TabEntry(nt,ETYP,n){
      funct=f;
    }
    ReturnNode(ProcedureNode *f, int nt, int t,char *n):TabEntry(nt,t,n){
      funct=f;
    }
    ReturnNode(ProcedureNode *f,int nt,int t,char *n,int l):TabEntry(nt,t,n,l){
      funct=f;
    }
    int getBeginIndex(){
      return beginIndex;
    }
    int getEndIndex(){
      return endIndex;
    }
    virtual LeftValueNode *getLeftValue(ExprNode *expr,int local);
    ProcedureNode *getFunction(){
      return funct;
    }
    virtual RightValueNode *getRightValue(ExprNode *expr,
		RealParametersNode *realPar,int local);
    virtual int getStackSize();
    void setType(ValueType *t){
      beginIndex=t->begin;
      endIndex=t->end;
      type=t->typ;
    }
    virtual void initArray(int local);
};

class ArrayNode;
class ArrayValNode: public RightValueNode{
  private:
    ArrayNode *val;
    ExprNode *index;
  public:
    ArrayValNode(ArrayNode *v, int t, ExprNode *i, int l):
		RightValueNode(t){
      val=v;
      index = i;
      local=l;
    }
    virtual int getStackSize();
    static int loadInstr(int type);
    virtual void generate();
};

class ArrayAdrNode: public LeftValueNode{
  private:
    ArrayNode *val;
    ExprNode *index;
  public:
    ArrayAdrNode(ArrayNode *v, int t, ExprNode *i, int l):
		LeftValueNode(ARRAYTYP, t){
      val=v;
      index = i;
      local=l;
    }
    virtual int getStackSize();
    virtual void generateBegin();
    virtual void generate();
    static int storeInstr(int type);
};

class ArrayNode: public TabEntry{
  private:
    int beginIndex;
    int endIndex;
  public:
    ArrayNode(int t,char *n, int b, int e):TabEntry(ARRAYTYP,t|ATYP,n){
      beginIndex=b;
      endIndex=e;
    } 
    ArrayNode(int t,char *n, int b, int e, int l):TabEntry(ARRAYTYP,t|ATYP,n){
      beginIndex=b;
      endIndex=e;
      line = l;      
    } 
    ArrayNode(int t,char *n, int l):TabEntry(ARRAYTYP,t|ATYP,n){
      beginIndex=0;
      endIndex=0;
      line = l;      
    } 
    int getBeginIndex(){
      return beginIndex;
    }
    int getEndIndex(){
      return endIndex;
    }
    virtual RightValueNode *getRightValue(ExprNode *expr,
		RealParametersNode *realPar,int local);
    virtual LeftValueNode *getLeftValue(ExprNode *expr,int local);
    virtual void generate();
    virtual void initArray(int local);
};


class BinaryNode: public ExprNode{
  private:
    int oper;
    ExprNode *loperand;
    ExprNode *poperand;
    void generateRel(int i);
  public:
    BinaryNode(int t):ExprNode(t){
    }
    BinaryNode(int o, ExprNode *l, ExprNode *p);
    BinaryNode(int o, ExprNode *l, ExprNode *p, int t);
    static BinaryNode* construct(int o, ExprNode *p1, ExprNode *p2);
    virtual int getStackSize();
    virtual void generate();
};

class UnaryNode: public ExprNode{
  private:
    int oper;
    ExprNode *operand;
  public:
    UnaryNode(int t){
      type=t;
    }
    UnaryNode(int o, ExprNode *p, int t):ExprNode(t){
      oper = o;
      operand = p;
    }
    UnaryNode(int o, ExprNode *p){
      oper = o;
      operand = p;
      type=p->getType();
    }
    static UnaryNode *construct(int o, ExprNode *p);
    virtual int getStackSize();
    virtual void generate();
};


class ConstNode: public RightValueNode{
  public:
    ConstNode():RightValueNode(ETYP){
    }
    ConstNode(int t):RightValueNode(t){
    }
};

class BoolConstNode: public ConstNode{
  private:
    int boolVal;
  public:
    BoolConstNode(int b):ConstNode(BTYP){
      boolVal=b;
    }
    virtual int getStackSize();
    virtual void generate();
};

class IntConstNode: public ConstNode{
  private:
    int intVal;
  public:
    IntConstNode(int i):ConstNode(ITYP){
      intVal=i;
    }
    virtual int getStackSize();
    virtual void generate();
};

class CharConstNode: public ConstNode{
  private:
    char charVal;
  public:
    CharConstNode(char ch):ConstNode(CTYP){
      charVal=ch;
    }
    virtual int getStackSize();
    virtual void generate();
};

class StrConstNode: public ConstNode{
  private:
    char *strVal;
  public:
    StrConstNode(char *s):ConstNode(STYP){
      strVal=s;
    }
    virtual int getStackSize();
    virtual void generate();
};

class CallNode: public RightValueNode{
  protected:
    ProcedureNode *procedure;
    RealParametersNode *parameters;
    ExprNode *index;
  public:
    CallNode(int t):RightValueNode(t){
    }
    CallNode(ProcedureNode *proc, RealParametersNode *param, ExprNode *i):
		RightValueNode(proc->getType()){
      procedure=proc;
      parameters=param;
      index=i;
    }
    CallNode(ProcedureNode *proc,int t,RealParametersNode *param,ExprNode *i):
		RightValueNode(t){
      procedure=proc;
      parameters=param;
      index=i;
    }
    ProcedureNode *getProcedure(){
      return procedure;
    }
    static CallNode *construct(char *name, RealParametersNode *param);
    virtual int getStackSize();
    virtual void generate();
};

class CallStmntNode: public StatementNode{
  private:
    CallNode *functCall;
  public:
    CallStmntNode(CallNode *f){
      functCall=f;
    }
    CallStmntNode(ProcedureNode *proc, RealParametersNode *param){
      functCall= proc->construct(param);
    }
    static CallStmntNode *construct(char *name, RealParametersNode *param){
      return new CallStmntNode(CallNode::construct(name,param));
    }
    static CallStmntNode *construct(ProcedureNode *funct,
                RealParametersNode *param){
      return new CallStmntNode(funct->construct(param));
    }
    virtual int getStackSize();
    virtual void generate();
};

class LetStmntNode: public StatementNode{
  private:
    LeftValueNode* address;
    ExprNode *expr;
  public:
    LetStmntNode(){
      expr=NULL;
      address=NULL;
    }
    LetStmntNode(LeftValueNode *a, ExprNode *e){
      address=a;
      expr=e;
    }
    static LetStmntNode *construct(LeftValueNode *adr, ExprNode *e);
    int getStackSize();
    virtual void generate();
};

class WhileStmntNode: public StatementNode{
  private:
    ExprNode *expr;
    StatementNode *stm;
  public:
    WhileStmntNode(){
      expr=NULL;
      stm=NULL;
    }
    WhileStmntNode(ExprNode *e, StatementNode *s){
      expr=e;
      stm = s;  
    }
    static WhileStmntNode *construct(ExprNode *e,StatementNode *s);
    virtual int getStackSize();
    virtual void generate();
};

class IfStmntNode: public StatementNode{
  private:
    ExprNode *expr;
    StatementNode *stm;
  public:
    IfStmntNode(){
      expr=NULL;
      stm=NULL;
    }
    IfStmntNode(ExprNode *e, StatementNode *s){
      expr=e;
      stm = s;  
    }
    static IfStmntNode *construct(ExprNode *e,StatementNode *s);
    int getStackSize();
    virtual void generate();
};

class IfElseStmntNode: public StatementNode{
  private:
    ExprNode *expr;
    StatementNode *stm1;
    StatementNode *stm2;
  public:
    IfElseStmntNode(){
      expr=NULL;
      stm1=NULL;
      stm2=NULL;
    }
    IfElseStmntNode(ExprNode *e, StatementNode *s1, StatementNode *s2){
      expr=e;
      stm1 = s1;
      stm2 = s2;
    }
    static IfElseStmntNode *construct(ExprNode *e,StatementNode *s1, 
		StatementNode *s2);
    int getStackSize();
    virtual void generate();
};

class RealParametersNode: public PasNode{
  private:
    Vector *params;
  public:
    RealParametersNode(Vector *p){
      params=p;
    }
    Vector *getParams(){
      return params;
    }
    int getStackSize();
    virtual void generate();
};

class BreakStmntNode: public StatementNode{
  public:
    int getStackSize();
    static BreakStmntNode* construct(int isWhile);
    virtual void generate();
};
class EmptyStmntNode: public StatementNode{
  public:
    int getStackSize();
    virtual void generate();
};

class ReturnStmntNode: public StatementNode{
  public:
    int getStackSize();
    virtual void generate();
};

