#include <stdlib.h>
#include <stream.h>
#include <FlexLexer.h>
#include "javacls/javacls.h"
#include "javacls/javainst.h"
#include "Vector.h"
#include "types.h"
#include "ValueType.h"
#include "PasNode.h"
#include "VarTable.h"
#include "PascalLexer.h"
#include "Compiler.h"
#include "StandardProc.h"

extern Compiler *compiler;
extern JavaCode *javaCode;

int WriteProc::compatible(RealParametersNode *real){
  Vector *v = real->getParams();
  int size = v->getSize();
  for(int i=0;i<size;i++){
    if( ( (((ExprNode *)v->elementAt(i)) ->getType()) &ATYP) !=0){
      return FALSE;
    }
  }
  return TRUE;
}
void WriteProc::generateReturn(){
  compiler->genError("Call of method WriteProc::generateReturn");
}

void WriteProc::generate(){
}

void WriteProc::generateCall(RealParametersNode *real){
  compiler->genError("Call of method WriteProc::generateCall");
}
void WriteProc::generatePrint(int typ){
  switch(typ){
    case ITYP:
      javaCode->genCall(INVOKESTATIC,"Methods","printInt","(I)V");
      break;
    case CTYP:
      javaCode->genCall(INVOKESTATIC,"Methods","printChar","(C)V");
      break;
    case BTYP:
      javaCode->genCall(INVOKESTATIC,"Methods","printBoolean","(Z)V");
      break;
    case STYP:
      javaCode->genCall(INVOKESTATIC,"Methods","printString",
		"(Ljava/lang/String;)V");
      break;
    default:
      compiler->genError("Unknown type in write method");
  }
}

void WritelnProc::generateReturn(){
  compiler->genError("Call of method WritelnProc::generateReturn");
}

void WritelnProc::generate(){
}

void WritelnProc::generateCall(RealParametersNode *real){
  compiler->genError("Call of method WritelnProc::generateCall");
}
CallNode *WriteProc::construct(RealParametersNode *param){
  if(compatible(param) == TRUE){
    return new WriteCallNode(this, param);
  }
  else{
    compiler->yyerrorShort();
    cerr<<"Procedure write can has only simple types in arguments"<<endl;
    return new CallNode(ETYP);
  }
}
CallNode *WritelnProc::construct(RealParametersNode *param){
  if(compatible(param) == TRUE){
    return new WritelnCallNode(this, param);
  }
  else{
    compiler->yyerrorShort();
    cerr<<"Procedure writeln can has only simple types in arguments"<<endl;
    return new CallNode(ETYP);
  }
}
void WriteCallNode::generate(){
  Vector *v = parameters->getParams();
  int size = v->getSize();
  for(int i=0;i<size;i++){
    ((ExprNode *)v->elementAt(i)) ->generate();
    ((WriteProc* )procedure)->generatePrint(((ExprNode *)v->elementAt(i)) 
		->getType());
  }
}
int WriteCallNode::getStackSize(){
  Vector *v = parameters->getParams();
  int size = v->getSize();

  int s=0;
  int s2;
  for(int i=0;i<size;i++){
    s2=((ExprNode *)v->elementAt(i)) ->getStackSize();
    s=(s2>s)?s2:s;
  }
  return s;
}
void WritelnCallNode::generate(){
  WriteCallNode::generate();
  javaCode->genCall(INVOKESTATIC,"Methods","println","()V");  
}



int LengthProc::compatible(RealParametersNode *real){
  Vector *v = real->getParams();
  int size = v->getSize();
  if(size!=1){
    return FALSE;
  }
  int t=((ExprNode *)v->elementAt(0)) ->getType();
  if( (t&ATYP)==0 && t!=STYP){
    return FALSE;
  }
  return TRUE;
}
void LengthProc::generateReturn(){
  compiler->genError("Call of method LengthProc::generateReturn");
}

void LengthProc::generate(){
}

void LengthProc::generateCall(RealParametersNode *real){
  if(((ExprNode*)real->getParams()->elementAt(0))->getType() == STYP)
  {  
    javaCode->genCall(INVOKESTATIC,"Methods","strlen","(Ljava/lang/String;)I");
  }
  else{
    javaCode->gen(ARRAYLENGTH);
  }
}
CallNode *LengthProc::construct(RealParametersNode *param){
  if(compatible(param) == TRUE){
    return new LengthCallNode(this, param);
  }
  else{
    compiler->yyerrorShort();
    cerr<<"Function lengt can has only array in argument"<<endl;
    return new CallNode(ETYP);
  }
}
void LengthCallNode::generate(){
  procedure->generateCall(parameters);
}
int LengthCallNode::getStackSize(){
  return 0;
}


int ReadIntProc::compatible(RealParametersNode *real){
  Vector *v = real->getParams();
  int size = v->getSize();
  if(size!=0){
    return FALSE;
  }
  return TRUE;
}
void ReadIntProc::generateReturn(){
  compiler->genError("Call of method ReadIntProc::generateReturn");
}

void ReadIntProc::generate(){
}

void ReadIntProc::generateCall(RealParametersNode *real){
  javaCode->genCall(INVOKESTATIC,"Methods","readInt","()I");
}
CallNode *ReadIntProc::construct(RealParametersNode *param){
  if(compatible(param) == TRUE){
    return new ReadIntCallNode(this, param);
  }
  else{
    compiler->yyerrorShort();
    cerr<<"Function readInt has not arguments"<<endl;
    return new CallNode(ETYP);
  }
}
void ReadIntCallNode::generate(){
  procedure->generateCall(parameters);
}
int ReadIntCallNode::getStackSize(){
  return 1;
}


int ReadCharProc::compatible(RealParametersNode *real){
  Vector *v = real->getParams();
  int size = v->getSize();
  if(size!=0){
    return FALSE;
  }
  return TRUE;
}
void ReadCharProc::generateReturn(){
  compiler->genError("Call of method ReadCharProc::generateReturn");
}

void ReadCharProc::generate(){
}

void ReadCharProc::generateCall(RealParametersNode *real){
  javaCode->genCall(INVOKESTATIC,"Methods","readChar","()C");
}
CallNode *ReadCharProc::construct(RealParametersNode *param){
  if(compatible(param) == TRUE){
    return new ReadCharCallNode(this, param);
  }
  else{
    compiler->yyerrorShort();
    cerr<<"Function readChar has not arguments"<<endl;
    return new CallNode(ETYP);
  }
}
void ReadCharCallNode::generate(){
  procedure->generateCall(parameters);
}
int ReadCharCallNode::getStackSize(){
  return 1;
}


int ReadBooleanProc::compatible(RealParametersNode *real){
  Vector *v = real->getParams();
  int size = v->getSize();
  if(size!=0){
    return FALSE;
  }
  return TRUE;
}
void ReadBooleanProc::generateReturn(){
  compiler->genError("Call of method ReadBooleanProc::generateReturn");
}

void ReadBooleanProc::generate(){
}

void ReadBooleanProc::generateCall(RealParametersNode *real){
  javaCode->genCall(INVOKESTATIC,"Methods","readBoolean","()Z");
}
CallNode *ReadBooleanProc::construct(RealParametersNode *param){
  if(compatible(param) == TRUE){
    return new ReadBooleanCallNode(this, param);
  }
  else{
    compiler->yyerrorShort();
    cerr<<"Function readBoolean has not arguments"<<endl;
    return new CallNode(ETYP);
  }
}
void ReadBooleanCallNode::generate(){
  procedure->generateCall(parameters);
}
int ReadBooleanCallNode::getStackSize(){
  return 1;
}


int ReadStringProc::compatible(RealParametersNode *real){
  Vector *v = real->getParams();
  int size = v->getSize();
  if(size!=0){
    return FALSE;
  }
  return TRUE;
}
void ReadStringProc::generateReturn(){
  compiler->genError("Call of method ReadStringProc::generateReturn");
}

void ReadStringProc::generate(){
}

void ReadStringProc::generateCall(RealParametersNode *real){
  javaCode->genCall(INVOKESTATIC,"Methods","readString",
	"()Ljava/lang/String;");
}
CallNode *ReadStringProc::construct(RealParametersNode *param){
  if(compatible(param) == TRUE){
    return new ReadStringCallNode(this, param);
  }
  else{
    compiler->yyerrorShort();
    cerr<<"Function readString has not arguments"<<endl;
    return new CallNode(ETYP);
  }
}
void ReadStringCallNode::generate(){
  procedure->generateCall(parameters);
}
int ReadStringCallNode::getStackSize(){
  return 1;
}


int EofProc::compatible(RealParametersNode *real){
  Vector *v = real->getParams();
  int size = v->getSize();
  if(size!=0){
    return FALSE;
  }
  return TRUE;
}
void EofProc::generateReturn(){
  compiler->genError("Call of method EofProc::generateReturn");
}

void EofProc::generate(){
}

void EofProc::generateCall(RealParametersNode *real){
  javaCode->genCall(INVOKESTATIC,"Methods","eof","()Z");
}
CallNode *EofProc::construct(RealParametersNode *param){
  if(compatible(param) == TRUE){
    return new EofCallNode(this, param);
  }
  else{
    compiler->yyerrorShort();
    cerr<<"Function eof has not arguments"<<endl;
    return new CallNode(ETYP);
  }
}
void EofCallNode::generate(){
  procedure->generateCall(parameters);
}
int EofCallNode::getStackSize(){
  return 1;
}


int ValProc::compatible(RealParametersNode *real){
  Vector *v = real->getParams();
  int size = v->getSize();
  if(size!=1){
    return FALSE;
  }
  int t=((ExprNode *)v->elementAt(0)) ->getType();
  if(t!=STYP){
    return FALSE;
  }
  return TRUE;
}
void ValProc::generateReturn(){
  compiler->genError("Call of method ValProc::generateReturn");
}

void ValProc::generate(){
}

void ValProc::generateCall(RealParametersNode *real){
  javaCode->genCall(INVOKESTATIC,"Methods","val","(Ljava/lang/String;)I");
}
CallNode *ValProc::construct(RealParametersNode *param){
  if(compatible(param) == TRUE){
    return new ValCallNode(this, param);
  }
  else{
    compiler->yyerrorShort();
    cerr<<"Function val can has only string in argument"<<endl;
    return new CallNode(ETYP);
  }
}
void ValCallNode::generate(){
  procedure->generateCall(parameters);
}
int ValCallNode::getStackSize(){
  return 1;
}
