/*
 * Autor: Martin Kot
 * Nazev: Prekladac
 * Studijni cislo: KOT119
 */

#define TRUE 1
#define FALSE 0
#include <iostream.h>
#include <FlexLexer.h>
#include <ctype.h>
#include "types.h"
#include "Vector.h"
#include "ValueType.h"
#include "PasNode.h"
#include "VarTable.h"
#include "PascalLexer.h"
#include "Compiler.h"
#include "StandardProc.h"

extern Compiler *compiler;

void VarTable::prodluz(){
  TabEntry **nove;
  int i;
  decl=decl+MINDECL;
  if((nove = new (TabEntry *)[decl])==NULL){
    cerr<<"Nedostatek pameti"<<endl;
    exit(0);
  }
  for(i=0;i<prvku;i++){
    nove[i]=vars[i];
  }
  delete vars;
  vars=nove;
}

VarTable::VarTable(){
  vars = new (TabEntry*)[MINDECL];
  prvku=0;
  decl=MINDECL;
}
void VarTable::addVar(TabEntry *var){
  if(prvku>=decl){
    prodluz();
  }
/*cerr<<"Add "<<var->getName()<<" "<<var->getNodeType()<<" "<<var->getType()<<endl;*/
  vars[prvku]=var;
  prvku++;
}

int VarTable::exist(char *n){
  int i;
  char *n2;
  if(n==NULL){
    return FALSE;
  }
  for(i=0;i<prvku;i++){
    n2=vars[i]->getName();
    if(n2==NULL){
      continue;
    }
    if(mstrcmp(n,n2)==0){
      return TRUE;
    }
  }
  return FALSE;
}

TabEntry *VarTable::getVal(char *n){
  int i;
  char *n2;
  if(n==NULL){
    return new TabEntry(ERRORTYP,ETYP);
  }
  for(i=0;i<prvku;i++){
    n2=vars[i]->getName();
    if(n2==NULL){
      continue;
    }
    if(mstrcmp(n,n2)==0){
      return vars[i];
    }
  }
/*cout<<"Nenasel "<<n<<endl;*/
/*printTable();*/
/*  cout<<endl<<endl;*/
  return new TabEntry(ERRORTYP,ETYP);
}

TabEntry *VarTable::elementAt(int i){
  return vars[i];
}

void VarTable::addVars(Vector *v){
  int p=v->getSize();
  int i;
  Variable *var;
  for(i=0;i<p;i++){
    var=(Variable *)v->elementAt(i);
    if(exist(var->getName())){
      compiler->error();
      cerr<<compiler->getLexer()->getFilename()<<var->getLine()<<": ";
      cerr<<"Identifier "<<var->getName()<<" is already used"<<endl;
      cerr<<"First definition at line "<<getVal(var->getName())->getLine();
      cerr<<endl;
    }
    else {
      addVar(var);
    }
  }
}
void VarTable::addVars(Vector *v,int param){
  int p=v->getSize();
  int i;
  Variable *var;
  for(i=0;i<p;i++){
    var=(Variable *)v->elementAt(i);
    if(exist(var->getName())){
      compiler->error();
      cerr<<compiler->getLexer()->getFilename()<<var->getLine()<<": ";
      cerr<<"Identifier "<<var->getName()<<" is already used"<<endl;
      cerr<<"First definition at line "<<getVal(var->getName())->getLine();
      cerr<<endl;
    }
    else {
      addVar(var);
      var->setParam(param);
    }
  }
}

void VarTable::printTable(){
  for(int i=0;i<prvku;i++){
    cout<<vars[i]->getName()<<" "<<vars[i]->getNodeType()<<" ";
    cout<<vars[i]->getType()<<" radek: "<<vars[i]->getLine()<<endl;
  }
}

int VarTable::mstrcmp(char *s1, char *s2){
  int l1=strlen(s1);
  int l2=strlen(s2);
  if(l1!=l2){
    return -1;
  }
  int i;
  for(i=0;i<l1;i++){
    if(toupper(s1[i])!=toupper(s2[i])){
      return -1;
    }
  }
  return 0;
}

int VarTable::localsSize(){
  int size=0;
  int i,j;
  TabEntry *retNode=NULL;
  j=0;
  for(int i=0;i<prvku;i++){
    if(vars[i]->getNodeType()==RETTYP || vars[i]->getNodeType()==RETPTYP){
      retNode=vars[i];
      j--;
    }
    else{
      vars[i]->setLocalAdr(j);
    }
    size+=PasNode::typesLength(vars[i]->getType());
    j++;
  }
  if(retNode!=NULL){
     retNode->setLocalAdr(j);
  }
  return size;
}
ReturnNode *VarTable::returnValue(){
  int i;
  for(i=0;i<prvku;i++){
    if(vars[i]->getNodeType()==RETTYP || vars[i]->getNodeType()==RETPTYP){
      return (ReturnNode *)vars[i];
    }
  }
  return NULL;
}

void VarTable::initArrays(int local){
  int i;
  for(i=0;i<prvku;i++){ 
    if(vars[i]->isParam()==FALSE){
      vars[i]->initArray(local);
    }
  }
}
void VarTable::addStandardMethods(){
  addVar(new WriteProc());
  addVar(new WritelnProc());  
  addVar(new LengthProc());  
  addVar(new ReadIntProc());  
  addVar(new ReadCharProc());  
  addVar(new ReadBooleanProc());  
  addVar(new ReadStringProc());  
  addVar(new EofProc());  
  addVar(new ValProc());  
}
