// Soubor: JavaCls.cpp
// Popis:  Implementace popisu tridy
// Autor:  benes@dcse.fee.vutbr.cz (Miroslav Benes)

# include "javaclsp.h"
# include <assert.h>
# include <stddef.h>

const char* mainName = "main";
extern const char* mainDescr = "([Ljava/lang/String;)V";

// 
// JavaClass
// =========

JavaClass::JavaClass()
{
   pp = NULL;
} // JavaClass::JavaClass

JavaClass::JavaClass(const char* aThisClass, const char* aSuperClass, unsigned short flags)
{
	pp = new JavaClassP(aThisClass, aSuperClass, flags | ACC_SUPER);
} // JavaClass::JavaClass

void JavaClass::create(const char* aThisClass, const char* aSuperClass, JavaUShort flags)
{
   delete pp;
   pp = new JavaClassP(aThisClass, aSuperClass, flags | ACC_SUPER);
} // JavaClass::create

bool JavaClass::load(const char* aThisClass)
{
   pp = JavaClassP::load(aThisClass);
   return( pp != NULL );
} // JavaClass::load

JavaClass::~JavaClass()
{
	delete pp;
} // JavaClass::~JavaClass

void JavaClass::setSourceFile(const char* s)
{
   assert(pp != NULL);
	pp->setSourceFile(s);
} // JavaClass::setSourceFile

const char* JavaClass::getClassName() const
{
   assert(pp != NULL);
   return pp->thisClass->str->s;
} // JavaClass::getClassName

JavaField* JavaClass::addField(const char* aName, const char* aDescr, unsigned short aFlags)
{
   assert(pp != NULL);
	return pp->addField(aName, aDescr, aFlags);
} // JavaClass::addField

JavaField* JavaClass::getField(const char* aName, const char* aDescr)
{
   assert(pp != NULL);
   return pp->getField(aName, aDescr);
} // JavaClass::getField

JavaMethod* JavaClass::addMethod(const char* aName, const char* aDescr, unsigned short aFlags)
{
   assert(pp != NULL);
   return pp->addMethod(aName, aDescr, aFlags);
} // JavaClass::addMethod

JavaMethod* JavaClass::getMethod(const char* aName, const char* aDescr)
{
   assert(pp != NULL);
   return pp->getMethod(aName, aDescr);
} // JavaClass::getMethod

void JavaClass::addDefaultInitMethod()
{
   JavaCode code(addMethod("<init>", "()V"));
   code.setLimits(1, 1);
   code.genLoad('L', 0);
   code.genCall(INVOKESPECIAL, "java/lang/Object", "<init>", "()V");
   code.gen(RETURN);
} // JavaClass::addDefaultInitMethod

void JavaClass::abort()
{
   assert(pp != NULL);
   pp->abort();
} // JavaClass::abort

bool JavaClass::save()
{
   assert(pp != NULL);
   return pp->save();
} // JavaClass::save

bool JavaClass::dump(const char* fileName, bool dumpCode, bool dumpConsts)
{
   assert(pp != NULL);
	return pp->dump(fileName, dumpCode, dumpConsts);
} // JavaClass::dump

const JavaConst* JavaClass::getConst(unsigned index)
{
   assert(pp != NULL);
   return pp->constPool[index];
} // JavaClass::getConst


//
// JavaField
// =========
JavaValue& JavaField::value()
{
   return ((JavaFieldP*)this)->currentValue;
} // JavaField::value


//
// JavaCode
// ========


JavaCode::JavaCode()
{
   pp = NULL;
} // JavaCode::JavaCode

JavaCode::JavaCode(JavaMethod* method)
{
   JavaMethodP* mp = (JavaMethodP*)method;
   if( mp->code == NULL )
      pp = new JavaCodeP(mp);
   else
      pp = mp->code;
} // JavaCode::JavaCode

JavaCode::~JavaCode()
{
   if( pp ) {
      pp->check();
   }
} // JavaCode::~JavaCode

void JavaCode::open(JavaMethod* method)
{
   assert(pp == NULL);

   JavaMethodP* mp = (JavaMethodP*)method;
   if( mp->code == NULL )
      pp = new JavaCodeP(mp);
   else
      pp = mp->code;
} // JavaCode::open

void JavaCode::close()
{
   if( pp != NULL ) {
      pp->check();
      pp = NULL;
   }
} // JavaCode::close

void JavaCode::setLimits(unsigned short aMaxStack, unsigned short aMaxLocals)
{
   pp->setLimits(aMaxStack, aMaxLocals);
} // JavaCode::setLimits

void JavaCode::getLimits(JavaUShort& aMaxStack, JavaUShort& aMaxLocals)
{
   aMaxStack = pp->maxStack;
   aMaxLocals = pp->maxLocals;
} // JavaCode::getLimits

void JavaCode::setPC(JavaUShort aPC)
{
   pp->currentPC = aPC;
} // JavaCode::setPC

JavaUShort JavaCode::getPC()
{
   return pp->currentPC;
} // JavaCode::getPC

JavaUShort JavaCode::getCodeSize()
{
   return pp->codeSize;
} // JavaCode::getCodeSize

JavaByte JavaCode::fetchByte()
{
   return pp->fetchByte();
} // JavaCode::fetchByte

JavaShort JavaCode::fetchShort()
{
   return pp->fetchShort();
} // JavaCode::fetchShort

JavaInt JavaCode::fetchInt()
{
   return pp->fetchInt();
} // JavaCode::fetchInt

# ifdef _MSC_VER
JavaLong JavaCode::fetchLong()
{
   return pp->fetchLong();
} // JavaCode::fetchLong
# endif

void JavaCode::gen(JavaInstr op)
{
   pp->genByte((JavaByte)op);
   if( op >= IRETURN && op <= RETURN )
      pp->accessible = false;
} // JavaCodeP::gen
#include "stream.h"
static char getTypeOffset(char type, bool all = false)
{
   switch( type ) {
      case JavaClass::TYPE_INT:
         return 0;
      case JavaClass::TYPE_LONG:
         return 1;
      case JavaClass::TYPE_FLOAT:
         return 2;
      case JavaClass::TYPE_DOUBLE:
         return 3;
      case JavaClass::TYPE_CLASS:
      case JavaClass::TYPE_ARRAY:
         return 4;
      case JavaClass::TYPE_BYTE:
         if( all ) return 5;
         break;
      case JavaClass::TYPE_CHAR:
         if( all ) return 6;
         break;
      case JavaClass::TYPE_SHORT:
         if( all ) return 7;
         break;
   }
cerr<<"M "<<type<<endl;
   assert(0);
   return 0;
} // getTypeOffset

void JavaCode::genLoad(char type, unsigned short index)
{
   char off = getTypeOffset(type);

   if( index < 4 )
      pp->genByte(ILOAD_0 + 4 * off + index);
   else if( index < 256 ) {
      pp->genByte(ILOAD + off);
      pp->genByte((JavaByte)index);
   } else {
      pp->genByte((JavaByte)WIDE);
      pp->genByte(ILOAD + off);
      pp->genShort(index);
   }
} // JavaCode::genLoad

void JavaCode::genStore(char type, unsigned short index)
{
   char off = getTypeOffset(type);

   if( index < 4 )
      pp->genByte(ISTORE_0 + 4 * off + index);
   else if( index < 256 ) {
      pp->genByte(ISTORE + off);
      pp->genByte((JavaByte)index);
   } else {
      pp->genByte((JavaByte)WIDE);
      pp->genByte(ISTORE + off);
      pp->genShort(index);
   }
} // JavaCode::genStore

void JavaCode::genIntConst(JavaInt val)
{
   if( val >= -1 && val <= 5 )
      pp->genByte((JavaByte)(ICONST_0 + val));
   else if( val >= -128 && val <= 127 ) {
      pp->genByte(BIPUSH);
      pp->genByte((JavaByte)val);
   } else if( val >= -32768l && val <= 32767 ) {
      pp->genByte(SIPUSH);
      pp->genShort((JavaShort)val);
   } else {
      JavaConstP* con = pp->method->parent->createIntegerConst(val);
      if( con->index <= 255 ) {
         pp->genByte(LDC);
         pp->genByte((JavaByte)con->index);
      } else {
         pp->genByte(LDC_W);
         pp->genShort(con->index);
      }
   }
} // JavaCode::genIntConst

void JavaCode::genStringConst(const char* val)
{
   JavaConstP* con = pp->method->parent->createStringConst(val);
   if( con->index <= 255 ) {
      pp->genByte(LDC);
      pp->genByte((JavaByte)con->index);
   } else {
      pp->genByte(LDC_W);
      pp->genShort(con->index);
   }
} // JavaCode::genStringConst

void JavaCode::genCall(JavaInstr op, const char* parent, const char* name, const char* descr)
{
   pp->genByte(op);
   JavaConstP* con = pp->method->parent->createRefConst(JavaConst::CONSTANT_Methodref, parent, name, descr);
   pp->genShort(con->index);
} // JavaCode::genCall

void JavaCode::genNew(const char* className)
{
   pp->genByte(NEW);
   JavaConstP* con = pp->method->parent->createClassConst(className);
   pp->genShort(con->index);
} // JavaCode::genNew

void JavaCode::genNewArray(const char* className)
{
   pp->genByte(ANEWARRAY);
   JavaConstP* con = pp->method->parent->createClassConst(className);
   pp->genShort(con->index);
} // JavaCode::genNewArray

void JavaCode::genNewArray(char type)
{
   pp->genByte(NEWARRAY);
   JavaByte tag;
   switch( type ) {
      case JavaClass::TYPE_BOOLEAN:   tag = 4;  break;
      case JavaClass::TYPE_CHAR:      tag = 5;  break;
      case JavaClass::TYPE_FLOAT:     tag = 6;  break;
      case JavaClass::TYPE_DOUBLE:    tag = 7;  break;
      case JavaClass::TYPE_BYTE:      tag = 8;  break;
      case JavaClass::TYPE_SHORT:     tag = 9;  break;
      case JavaClass::TYPE_INT:       tag = 10; break;
      case JavaClass::TYPE_LONG:      tag = 11; break;
      default:
         assert(0);
         break;
   }
   pp->genByte(tag);
} // JavaCode::genNewArray

void JavaCode::genJump(JavaInstr op, JavaLabel label)
{
   JavaLabelP* lp = pp->labels[label];

   pp->genByte(op);
   if( lp->offset == JavaUndefLabel ) {
      pp->genShort(lp->lastRef);
      lp->lastRef = pp->currentPC - 2;
   } else
      pp->genShort(lp->offset - (pp->currentPC-1));
   if( op == GOTO )
      pp->accessible = false;
} // JavaCode::genJump

void JavaCode::genLabel(JavaLabel label)
{
   pp->defLabel(label, pp->currentPC);
   pp->accessible = true;
} // JavaCode::genLabel

JavaLabel JavaCode::getLabel()
{
   return pp->newLabel();
} // JavaCode::getLabel

bool JavaCode::isAccessible()
{
   return pp->accessible;
} // JavaCode::isAccessible

void JavaCode::genField(JavaInstr op, const char* parent, const char* name, const char* descr)
{
   pp->genByte(op);
   JavaConstP* con = pp->method->parent->createRefConst(JavaConst::CONSTANT_Fieldref, parent, name, descr);
   pp->genShort(con->index);
} // JavaCode::genField


//
// JavaConst
// =========

JavaConst::JavaConstTag JavaConst::getTag() const
{
   return ((JavaConstP*)this)->tag;
} // JavaConst::getTag

JavaCString JavaConst::getString() const
{
   JavaConstP* cp = (JavaConstP*)this;
   assert( cp->tag == CONSTANT_Utf8 || cp->tag == CONSTANT_String );
   return cp->tag == CONSTANT_Utf8 ? cp->s : cp->str->s;
} // JavaConst::getString

JavaInt JavaConst::getInteger() const
{
   JavaConstP* cp = (JavaConstP*)this;
   assert( cp->tag == CONSTANT_Integer );
   return cp->i;
} // JavaConst::getInteger

void JavaConst::getRef(JavaCString &cls, JavaCString &name, JavaCString &descr) const
{
   JavaConstP* cp = (JavaConstP*)this;
   assert( cp->tag == CONSTANT_Fieldref || cp->tag == CONSTANT_Methodref || cp->tag == CONSTANT_InterfaceMethodref );
   cls = cp->ref.cls->str->s;
   name = cp->ref.nt->nt.name->s;
   descr = cp->ref.nt->nt.type->s;
} // JavaConst::getRef

