// Soubor: JavaCls.h
// Popis:  Rozhran pro reprezentaci tdy
// Autor:  benes@dcse.fee.vutbr.cz (Miroslav Bene)

# ifndef _JAVACLS_H_
# define _JAVACLS_H_

# ifdef __BORLANDC__
typedef unsigned char bool;
const bool false = 0;
const bool true  = 1;
# endif

# include "javainst.h"

extern const char* mainName;
extern const char* mainDescr;

// zkladn datov typy
typedef signed char  JavaByte;
typedef short        JavaShort;
typedef long         JavaInt;
typedef float        JavaFloat;
typedef double       JavaDouble;
typedef short        JavaChar;

# ifdef _MSC_VER
typedef __int64      JavaLong;
# endif

typedef unsigned char	JavaUByte;
typedef unsigned short	JavaUShort;
typedef unsigned long	JavaUInt;
typedef unsigned short  JavaLabel;
typedef const char*     JavaCString;

class JavaClassP;
class JavaCodeP;
class JavaField;
class JavaMethod;
class JavaCode;
class JavaConst;

const JavaLabel JavaUndefLabel = 0xFFFF;

union JavaValue {
   JavaInt        i;
# ifdef _MSC_VER
   JavaLong       l;
# endif
   JavaFloat      f;
   JavaDouble     d;
};


// reprezentace tdy
class JavaClass {
public:
   
   // pstupov atributy
   enum {
      ACC_PUBLIC        = 0x0001, 
      ACC_PRIVATE       = 0x0002, 
      ACC_PROTECTED     = 0x0004, 
      ACC_STATIC        = 0x0008,
      ACC_FINAL         = 0x0010,
      ACC_SYNCHRONIZED  = 0x0020, 
      ACC_SUPER         = 0x0020,
      ACC_VOLATILE      = 0x0040, 
      ACC_TRANSIENT     = 0x0080,
      ACC_NATIVE        = 0x0100, 
      ACC_INTERFACE     = 0x0200,
      ACC_ABSTRACT      = 0x0400,
   };
   
   // datov typy
   enum {
      TYPE_BYTE         = 'B',
      TYPE_CHAR         = 'C',
      TYPE_DOUBLE       = 'D',
      TYPE_FLOAT        = 'F',
      TYPE_INT          = 'I',
      TYPE_LONG         = 'J',
      TYPE_CLASS        = 'L',
      TYPE_SHORT        = 'S',
      TYPE_BOOLEAN      = 'Z',
      TYPE_ARRAY        = '[',
      TYPE_VOID         = 'V',
   };
   
   // default konstruktor
   JavaClass();
   
   // konstruktor
   JavaClass(const char* aThisClass, const char* aSuperClass = "java/lang/Object",
      JavaUShort flags = ACC_PUBLIC);
   
   // destruktor
   ~JavaClass();
   
   // Nastaven jmna zdrojovho souboru
   void setSourceFile(const char* s);
   
   // zskn jmna tdy
   const char* getClassName() const;

   // deklarace poloky tdy
   JavaField* addField(const char* aName, const char* aDescr, JavaUShort aFlags = ACC_PUBLIC);
   
   // vyhledn poloky tdy
   JavaField* getField(const char* aName, const char* aDescr);

   // deklarace metody
   JavaMethod* addMethod(const char* aName, const char* aDescr, JavaUShort aFlags = ACC_PUBLIC);

   // vyhledani metody
   JavaMethod* getMethod(const char* aName, const char* aDescr);

   // deklarace metody <init>
   void JavaClass::addDefaultInitMethod();

   // zskn konstanty
   const JavaConst* getConst(unsigned index);
   
   // vytvoen nov tdy (pouze ve spojen s default konstruktorem)
   void create(const char* aThisClass, const char* aSuperClass = "java/lang/Object",
      JavaUShort flags = ACC_PUBLIC);
   
   // ten ze souboru .class
   bool load(const char* aThisClass);
   
   // uloen do souboru .class
   bool save();

   // zruseni souboru .class
   void abort();

   // ladic vpis
   bool dump(const char* fileName = 0, bool dumpCode = true, bool dumpConsts = false);
   
private:
   JavaClassP*	pp;
}; // JavaClass

// reprezentace konstanty
class JavaConst {
public:

   enum JavaConstTag {
      CONSTANT_Unknown  = 0, 
      CONSTANT_Utf8     = 1,
      CONSTANT_Unicode  = 2,
      CONSTANT_Integer  = 3,
      CONSTANT_Float    = 4,
      CONSTANT_Long     = 5,
      CONSTANT_Double   = 6,
      CONSTANT_Class    = 7,
      CONSTANT_String   = 8,
      CONSTANT_Fieldref = 9,
      CONSTANT_Methodref = 10,
      CONSTANT_InterfaceMethodref = 11,
      CONSTANT_NameAndType = 12,
   };

   JavaConstTag getTag() const;
   JavaCString getString() const;
   void getRef(JavaCString &cls, JavaCString &name, JavaCString &descr) const;
   JavaInt getInteger() const;
protected:
   JavaConst() {}
}; // JavaConst

// reprezentace metody
class JavaMethod {
protected:
   JavaMethod() {}
}; // JavaMethod

// reprezentace poloky
class JavaField {
public:
   JavaValue& value();
protected:
   JavaField() {}
}; // JavaField


// reprezentace kdu metody 
class JavaCode {
public:
   // konstruktory
   JavaCode();
   JavaCode(JavaMethod* method);
   
   // destruktor
   ~JavaCode();
   
   // zatek generovn (v ppad vcensobnho pouit objektu JavaCode)
   void open(JavaMethod* method);
   
   // konec generovn (v ppad vcensobnho pouit objektu JavaCode)
   void close();
   
   // nastaven limit
   void setLimits(JavaUShort aMaxStack, JavaUShort aMaxLocals);

   // ziskani hodnoty limitu
   void getLimits(JavaUShort& aMaxStack, JavaUShort& aMaxLocals);

   // jednoduch instrukce
   void gen(JavaInstr op);

   // lokln promnn
   void genLoad(char type, JavaUShort index);
   void genStore(char type, JavaUShort index);

   // globln promnn
   void genField(JavaInstr op, const char* parent, const char* name, const char* descr);

   // konstanty
   void genIntConst(JavaInt val);
   void genStringConst(const char* val);

   // voln metod
   void genCall(JavaInstr op, const char* parent, const char* name, const char* descr);

   // new
   void genNew(const char* className);
   void genNewArray(const char* className);
   void genNewArray(char simpleType);

   // nvt a skoky
   JavaLabel getLabel();
   void genJump(JavaInstr op, JavaLabel label);
   void genLabel(JavaLabel label);
   bool isAccessible();          // je dal instrukce dostupn?

   void setPC(JavaUShort aPC);
   JavaUShort getPC();
   JavaUShort getCodeSize();
   JavaByte  fetchByte();
   JavaShort fetchShort();
   JavaInt   fetchInt();
# ifdef _MSC_VER
   JavaLong  fetchLong();
# endif

private:
   JavaCodeP* pp;
}; // JavaCode

# endif
