// Soubor: JavaClsP.cpp
// Popis:  Implementace privatnich metod popisu tridy
// Autor:  benes@dcse.fee.vutbr.cz (Miroslav Benes)

# include <stdio.h>
# include <stdlib.h>
# include <string.h>
# include <assert.h>
# include "javaclsp.h"

# ifndef _MAX_PATH
# define _MAX_PATH 255
# endif

// JavaFieldP - polozka tabulky promennych tridy
// =============================================

JavaFieldP::JavaFieldP(JavaClassP* aParent)
{
   parent = aParent;
   memset(&currentValue, 0, sizeof(currentValue));
} // JavaFieldP::JavaFieldP

JavaFieldP::JavaFieldP(JavaClassP* aParent, JavaConstP* aName, JavaConstP* aDescr, unsigned short flags)
: name(aName), descriptor(aDescr), accessFlags(flags) 
{
   parent = aParent;
   constantValue = NULL;
   memset(&currentValue, 0, sizeof(currentValue));
} // JavaFieldP::JavaFieldP


// JavaMethodP - polozka tabulky metod
// ===================================

JavaMethodP::JavaMethodP(JavaClassP* aParent)
{
   parent = aParent;
} // JavaMethodP::JavaMethodP

JavaMethodP::JavaMethodP(JavaClassP* aParent, JavaConstP* aName, JavaConstP* aDescr, unsigned short flags)
: name(aName), descriptor(aDescr), accessFlags(flags) 
{
   parent = aParent;
   code = NULL;
} // JavaMethodP::JavaMethodP


// JavaClassP - popis tridy
// ========================

JavaClassP::JavaClassP()
{
   thisClass = NULL;
   superClass = NULL;
   accessFlags = 0;
   sourceFile = NULL;
   
   sourceFileAttr = NULL;
   constantValueAttr = NULL;
   codeAttr = NULL;
} // JavaClassP::JavaClassP

JavaClassP::JavaClassP(const char* aThisClass, const char* aSuperClass, unsigned short flags)
{
   addConst(NULL);
   codeAttr          = createUtf8Const("Code");
   sourceFileAttr    = createUtf8Const("SourceFile");
   constantValueAttr = createUtf8Const("ConstantValue");
   exceptionsAttr    = createUtf8Const("Exceptions");
   
   thisClass   = createClassConst(aThisClass);
   superClass  = createClassConst(aSuperClass);
   accessFlags = flags;
   
   sourceFile  = NULL;
} // JavaClassP::JavaClassP

JavaClassP::~JavaClassP()
{
} // JavaClassP::~JavaClassP

void JavaClassP::setSourceFile(const char* s)
{
   sourceFile     = createUtf8Const(s);
} // setSourceFile

JavaFieldP* JavaClassP::addField(const char* aName, const char* aDescr, unsigned short flags)
{
   JavaFieldP* field = new JavaFieldP(this, createUtf8Const(aName), createUtf8Const(aDescr), flags);
   fields.insertLast(field);
   return field;
} // JavaClassP::addField

JavaField* JavaClassP::getField(const char* aName, const char* aDescr)
{
   for(int i = 0; i < fields.getCount(); i++) {
      JavaFieldP* field = fields[i];
      if( strcmp(field->name->s, aName) == 0 && strcmp(field->descriptor->s, aDescr) == 0 )
         return field;
   }
   return NULL;
} // JavaClassP::getField

JavaMethodP* JavaClassP::addMethod(const char* aName, const char* aDescr, unsigned short flags)
{
   JavaMethodP* method = new JavaMethodP(this, createUtf8Const(aName), createUtf8Const(aDescr), flags);
   methods.insertLast(method);
   return method;
} // JavaClassP::addField

JavaMethod* JavaClassP::getMethod(const char* aName, const char* aDescr)
{
   for(int i = 0; i < methods.getCount(); i++) {
      JavaMethodP* method = methods[i];
      if( strcmp(method->name->s, aName) == 0 && strcmp(method->descriptor->s, aDescr) == 0 )
         return method;
   }
   return NULL;
} // JavaClassP::getMethod

void JavaClassP::addConst(JavaConstP* cp)
{
   if( cp != NULL )
      cp->index = constPool.getCount();
   constPool.insertLast(cp);
} // JavaClassP::addConst

JavaConstP* JavaClassP::createUtf8Const(const char* s)
{
   for(unsigned i = 1; i < constPool.getCount(); i++) {
      JavaConstP* cp = constPool[i];
      if( cp->tag == JavaConst::CONSTANT_Utf8 && strcmp(cp->s, s) == 0 )
         return cp;
   }
   
   JavaConstP* cp = new JavaConstP(JavaConst::CONSTANT_Utf8);
   cp->s = new char[strlen(s)+1];
   strcpy(cp->s, s);
   addConst(cp);
   return cp;
} // JavaConstP::createUtf8Const

JavaConstP*	JavaClassP::createIntegerConst(JavaInt val)
{
   for(unsigned i = 1; i < constPool.getCount(); i++) {
      JavaConstP* cp = constPool[i];
      if( cp->tag == JavaConst::CONSTANT_Integer && cp->i == val )
         return cp;
   }
   
   JavaConstP* cp = new JavaConstP(JavaConst::CONSTANT_Integer);
   cp->i = val;
   addConst(cp);
   return cp;
} // JavaClassP::createIntegerConst

JavaConstP* JavaClassP::createStringConst(const char* val)
{
   JavaConstP* stringVal = createUtf8Const(val);

   for(unsigned i = 1; i < constPool.getCount(); i++) {
      JavaConstP* cp = constPool[i];
      if( cp->tag == JavaConst::CONSTANT_String && cp->str == stringVal )
         return cp;
   }
   
   JavaConstP* cp = new JavaConstP(JavaConst::CONSTANT_String);
   cp->str = stringVal;
   addConst(cp);
   return cp;
} // JavaClassP::createStringConst

JavaConstP* JavaClassP::createClassConst(const char* cls)
{
   JavaConstP* className = createUtf8Const(cls);
   
   for(unsigned i = 1; i < constPool.getCount(); i++) {
      JavaConstP* cp = constPool[i];
      if( cp->tag == JavaConst::CONSTANT_Class && cp->str == className )
         return cp;
   }
   
   JavaConstP* cp = new JavaConstP(JavaConst::CONSTANT_Class);
   cp->str = className;
   addConst(cp);
   return cp;
} // JavaConstP::createClassConst

JavaConstP* JavaClassP::createNameAndTypeConst(const char* aName, const char* aType)
{
   JavaConstP* name = createUtf8Const(aName);
   JavaConstP* type = createUtf8Const(aType);
   
   for(unsigned i = 1; i < constPool.getCount(); i++) {
      JavaConstP* cp = constPool[i];
      if( cp->tag == JavaConst::CONSTANT_NameAndType && cp->nt.name == name && cp->nt.type == type )
         return cp;
   }
   
   JavaConstP* cp = new JavaConstP(JavaConst::CONSTANT_NameAndType);
   cp->nt.name = name;
   cp->nt.type = type;
   addConst(cp);
   return cp;
} // JavaClassP::createNameAndTypeConst

JavaConstP* JavaClassP::createRefConst(JavaConst::JavaConstTag tag, const char* cls, const char* meth, const char* descr)
{
   JavaConstP* nt = createNameAndTypeConst(meth, descr);
   JavaConstP* cc = createClassConst(cls);
   
   for(unsigned i = 1; i < constPool.getCount(); i++) {
      JavaConstP* cp = constPool[i];
      if( cp->tag == tag && cp->ref.cls == cc && cp->ref.nt == nt )
         return cp;
   }
   
   JavaConstP* cp = new JavaConstP(tag);
   cp->ref.cls = cc;
   cp->ref.nt = nt;
   addConst(cp);
   return cp;
} // JavaClassP::createRefConst


// JavaCodeP - cilovy kod metody
// =============================

JavaCodeP::JavaCodeP(JavaMethodP* aMethod)
{
   method = aMethod;
   aMethod->code = this;
   maxStack = maxLocals = 0;
   currentPC = 0;
   maxCodeSize = 0;
   codeSize = 0;
   code = NULL;
   accessible = true;
} // JavaCodeP::JavaCodeP

void JavaCodeP::setLimits(unsigned short aMaxStack, unsigned short aMaxLocals)
{
   maxStack = aMaxStack;
   maxLocals = aMaxLocals;
} // JavaCodeP::setLimits


void JavaCodeP::setCode(unsigned char* buf, unsigned short size) 
{
   assert( this != NULL );
   code = buf;
   codeSize = maxCodeSize = size;
   currentPC = 0;
}; // JavaCodeP::setCode

void JavaCodeP::genData(void* buf, JavaUShort size)
{
   assert( this != NULL );
   const unsigned increment = 64;
   if( currentPC + size > maxCodeSize ) {
      unsigned char* newCode = new unsigned char[maxCodeSize + increment];
      memcpy(newCode, code, codeSize);
      delete code;
      code = newCode;
      maxCodeSize += increment;
   }
   memcpy(code + currentPC, buf, size);
   currentPC += size;
   if( currentPC > codeSize )
      codeSize = currentPC;
} // JavaCodeP::genData

void JavaCodeP::genByte(JavaUByte x)
{
   genData(&x, 1);
} // JavaCodeP::genUnsignedByte

void JavaCodeP::genShort(JavaShort x)
{
   unsigned char b[2];
   b[0] = (unsigned char)(x >> 8);
   b[1] = (unsigned char)(x >> 0);
   genData(b, 2);
} // JavaCodeP::genUnsignedWord

void JavaCodeP::genInt(JavaInt x)
{
   unsigned char b[4];
   b[0] = (unsigned char)(x >> 24);
   b[1] = (unsigned char)(x >> 16);
   b[2] = (unsigned char)(x >>  8);
   b[3] = (unsigned char)(x >>  0);
   genData(b, 4);
} // JavaCodeP::genInt

# ifdef _MSC_VER
void JavaCodeP::genLong(JavaLong x)
{
   unsigned char b[8];
   b[0] = (unsigned char)(x >> 56);
   b[1] = (unsigned char)(x >> 48);
   b[2] = (unsigned char)(x >> 40);
   b[3] = (unsigned char)(x >> 32);
   b[4] = (unsigned char)(x >> 24);
   b[5] = (unsigned char)(x >> 16);
   b[6] = (unsigned char)(x >>  8);
   b[7] = (unsigned char)(x >>  0);
   genData(b, 8);
} // JavaCodeP::genInt
# endif

JavaUInt JavaCodeP::getAttrLength() const
{
   JavaUInt len = 12 + codeSize;
   
   // len += 8 * exceptionTableLength;
   // len += lineNumberTableLength
   // len += localVariableTableLength
   return len;
} // JavaCodeP::getAttrLength

void JavaCodeP::fetchData(void* ptr, size_t size)
{
   assert(currentPC+size <= codeSize);
   memcpy(ptr, &code[currentPC], size);
   currentPC += size;
} // JavaCodeP::fetchData

JavaByte JavaCodeP::fetchByte()
{
   assert(currentPC < codeSize);
   JavaByte x = code[currentPC++];
   return x;
} // JavaCodeP::fetchByte

JavaShort JavaCodeP::fetchShort()
{
   JavaUByte b[2];
   fetchData(b, 2);
   
   JavaShort x = ((unsigned short)b[0] << 8) | b[1];
   return x;
} // JavaCodeP::fetchShort

JavaInt JavaCodeP::fetchInt()
{
   JavaUByte b[4];
   fetchData(b, 4);
   
   JavaInt x = (((((((long)b[0] << 8) | b[1]) << 8) | b[2])) << 8) | b[3];
   return x;
} // JavaCodeP::fetchInt


# ifdef _MSC_VER
JavaLong JavaCodeP::fetchLong()
{
   JavaUByte b[8];
   fetchData(b, 8);
   
   JavaInt x = b[0];
   x = (x << 8) | b[1];
   x = (x << 8) | b[2];
   x = (x << 8) | b[3];
   x = (x << 8) | b[4];
   x = (x << 8) | b[5];
   x = (x << 8) | b[6];
   x = (x << 8) | b[7];
   return x;
} // JavaCodeP::fetchInt
# endif

// GENEROVANI CILOVEHO SOUBORU
// ===========================

static FILE* f;

static void writeData(void* ptr, unsigned short size)
{
   fwrite(ptr, 1, size, f);
} // writeData

static void writeUnsignedByte(unsigned char x)
{
   putc(x, f);
} // writeUnsignedByte

static void writeUnsignedWord(unsigned short x)
{
   unsigned char b[2];
   b[0] = (x >> 8) & 0xff;
   b[1] = (x >> 0) & 0xff;
   writeData(b, 2);
} // writeUnsignedWord

static void writeInt(long x)
{
   unsigned char b[4];
   b[0] = (x >> 24) & 0xff;
   b[1] = (x >> 16) & 0xff;
   b[2] = (x >>  8) & 0xff;
   b[3] = (x >>  0) & 0xff;
   writeData(b, 4);
} // writeInt

void JavaCodeP::write()
{
   writeInt(getAttrLength());
   writeUnsignedWord(maxStack);
   writeUnsignedWord(maxLocals);
   writeInt(codeSize);
   writeData(code, codeSize);
   writeUnsignedWord(0);						// pocet polozek tabulky vyjimek
   writeUnsignedWord(0);						// pocet atributu
} // JavaCodeP::write

void JavaClassP::writeConst(JavaConstP* cp)
{
   writeUnsignedByte(cp->tag);
   switch( cp->tag ) {
   case JavaConst::CONSTANT_Utf8: {
      unsigned short len = strlen(cp->s);
      writeUnsignedWord(len);
      writeData(cp->s, len);
      break;
   }
   case JavaConst::CONSTANT_Integer:
      writeInt(cp->i);
      break;
   case JavaConst::CONSTANT_Class:
   case JavaConst::CONSTANT_String:
      writeUnsignedWord(cp->str->index);
      break;
   case JavaConst::CONSTANT_Methodref:
   case JavaConst::CONSTANT_InterfaceMethodref:
   case JavaConst::CONSTANT_Fieldref:
      writeUnsignedWord(cp->ref.cls->index);
      writeUnsignedWord(cp->ref.nt->index);
      break;
   case JavaConst::CONSTANT_NameAndType:
      writeUnsignedWord(cp->nt.name->index);
      writeUnsignedWord(cp->nt.type->index);
      break;
   default:
      assert(0);
      break;
   }
} // JavaClassP::writeConst

void JavaClassP::writeField(JavaFieldP* fp)
{
   writeUnsignedWord(fp->accessFlags);
   writeUnsignedWord(fp->name->index);
   writeUnsignedWord(fp->descriptor->index);
   if( fp->constantValue ) {
      writeUnsignedWord(1);
      writeUnsignedWord(constantValueAttr->index);
      writeInt(2);
      writeUnsignedWord(fp->constantValue->index);
   } else
      writeUnsignedWord(0);
} // JavaClassP::writeField

void JavaClassP::writeMethod(JavaMethodP* mp)
{
   writeUnsignedWord(mp->accessFlags);
   writeUnsignedWord(mp->name->index);
   writeUnsignedWord(mp->descriptor->index);
   if( mp->code ) {
      writeUnsignedWord(1);					// pocet atributu
      writeUnsignedWord(codeAttr->index);
      mp->code->write();
   } else
      writeUnsignedWord(0);
} // JavaClassP::writeMethod

void JavaClassP::abort()
{
   char fileName[_MAX_PATH];
   sprintf(fileName, "%s.class", thisClass->str->s);
   remove(fileName);
}

bool JavaClassP::save()
{
   unsigned i;

   char fileName[_MAX_PATH];
   sprintf(fileName, "%s.class", thisClass->str->s);

   f = fopen(fileName, "wb");
   if( f == NULL )
      return false;

   // zhlav
   writeInt(JavaClassMagicNumber);
   writeUnsignedWord(JavaClassMinorVersion);
   writeUnsignedWord(JavaClassMajorVersion);

   // tabulka konstant
   writeUnsignedWord(constPool.getCount());
   for(i = 1; i < constPool.getCount(); i++)
      writeConst(constPool[i]);
   
   // pstupov prva
   writeUnsignedWord(accessFlags);
   
   // aktuln a bzov tda
   writeUnsignedWord(thisClass->index);
   writeUnsignedWord(superClass->index);
   
   // seznam rozhran
   writeUnsignedWord(interfaces.getCount());
   for(i = 0; i < interfaces.getCount(); i++)
      writeUnsignedWord(interfaces[i]->index);
   
   // seznam poloek
   writeUnsignedWord(fields.getCount());
   for(i = 0; i < fields.getCount(); i++)
      writeField(fields[i]);
   
   // seznam metod
   writeUnsignedWord(methods.getCount());
   for(i = 0; i < methods.getCount(); i++)
      writeMethod(methods[i]);
   
   // seznam atribut
   if( sourceFile ) {
      writeUnsignedWord(1);								// pocet atributu
      writeUnsignedWord(sourceFileAttr->index);		// "SourceFile"
      writeInt(2);											// delka atributu = 2
      writeUnsignedWord(sourceFile->index);			// atribut
   } else
      writeUnsignedWord(0);
   
   fclose(f);
   return true;
} // JavaClassP::save


// LADIC VPIS TDY
// ==================

static const char* instrName[] = {
   "nop",         "aconst_null", "iconst_m1",   "iconst_0",    "iconst_1",    "iconst_2", 
   "iconst_3",    "iconst_4",    "iconst_5",    "lconst_0",    "lconst_1",    "fconst_0",
   "fconst_1",    "fconst_2",    "dconst_0",    "dconst_1",    "bipush",      "sipush",
   "ldc",         "ldc_w",       "ldc2w",       "iload",       "lload",       "fload",
   "dload",       "aload",       "iload_0",     "iload_1",     "iload_2",     "iload_3",
   "lload_0",     "lload_1",     "lload_2",     "lload_3",     "fload_0",     "fload_1",
   "fload_2",     "fload_3",     "dload_0",     "dload_1",     "dload_2",     "dload_3",
   "aload_0",     "aload_1",     "aload_2",     "aload_3",     "iaload",      "laload",
   "faload",      "daload",      "aaload",      "baload",      "caload",      "saload",
   "istore",      "lstore",      "fstore",      "dstore",      "astore",      "istore_0",
   "istore_1",    "istore_2",    "istore_3",    "lstore_0",    "lstore_1",    "lstore_2",
   "lstore_3",    "fstore_0",    "fstore_1",    "fstore_2",    "fstore_3",    "dstore_0",
   "dstore_1",    "dstore_2",    "dstore_3",    "astore_0",    "astore_1",    "astore_2",
   "astore_3",    "iastore",     "lastore",     "fastore",     "dastore",     "aastore",
   "bastore",     "castore",     "sastore",     "pop",         "pop2",        "dup",
   "dup_x1",      "dup_x2",      "dup2",        "dup2_x1",     "dup2_x2",     "swap",
   "iadd",        "ladd",        "fadd",        "dadd",        "isub",        "lsub",
   "fsub",        "dsub",        "imul",        "lmul",        "fmul",        "dmul",
   "idiv",        "ldiv",        "fdiv",        "ddiv",        "irem",        "lrem",
   "frem",        "drem",        "ineg",        "lneg",        "fneg",        "dneg",
   "ishl",        "lshl",        "ishr",        "lshr",        "iushr",       "lushr",
   "iand",        "land",        "ior",         "lor",         "ixor",        "lxor",
   "iinc",        "i2l",         "i2f",         "i2d",         "l2i",         "l2f",
   "l2d",         "f2i",         "f2l",         "f2d",         "d2i",         "d2l",
   "d2f",         "int2byte",    "int2char",    "int2short",   "lcmp",        "fcmpl",
   "fcmpg",       "dcmpl",       "dcmpg",       "ifeq",        "ifne",        "iflt",
   "ifge",        "ifgt",        "ifle",        "if_icmpeq",   "if_icmpne",   "if_icmplt",
   "if_icmpge",   "if_icmpgt",   "if_icmple",   "if_acmpeq",   "if_acmpne",   "goto",
   "jsr",         "ret",         "tableswitch", "lookupswitch","ireturn",     "lreturn",
   "freturn",     "dreturn",     "areturn",     "return",      "getstatic",   "putstatic",
   "getfield",    "putfield",    "invokevirtual","invokespecial","invokestatic","invokeinterface",
   "?186?",       "new",         "newarray",    "anewarray",   "arraylength", "athrow",
   "checkcast",   "instanceof",  "monitorenter","monitorexit", "wide",        "multianewarray",
   "ifnull",      "ifnonnull",   "goto_w",      "jsr_w",       "breakpoint",  "?203",
   "?204?",       "?205?",       "?206?",       "?207?",       "?208?",       "ret_w",
};

static const char* typeName[] = {
   "boolean", "char", "float", "double", "byte", "short", "int", "long"
};

static const char* dumpClassName(FILE* f, const char* descr)
{
   char ch;
   while( (ch = *descr) != '\0' && ch != ';' && ch != ')' ) {
      if( ch == '/' )
         putc('.', f);
      else
         putc(ch, f);
      descr++;
   }
   return descr;
} // dumpClassName

static const char* dumpType(FILE* f, const char* descr)
{
   switch( *descr++ ) {
      case 'B': 
         fprintf(f, "byte");
         break;
      case 'C': 
         fprintf(f, "char");
         break;
      case 'D': 
         fprintf(f, "double");
         break;
      case 'F': 
         fprintf(f, "float");
         break;
      case 'I': 
         fprintf(f, "int");
         break;
      case 'J': 
         fprintf(f, "long");
         break;
      case 'S': 
         fprintf(f, "short");
         break;
      case 'Z': 
         fprintf(f, "boolean");
         break;
      case 'V': 
         fprintf(f, "void");
         break;
      case 'L': {
         descr = dumpClassName(f, descr);
         assert(*descr == ';');
         descr++;
         break;
      }
      case '[': {
         descr = dumpType(f, descr);
         fprintf(f, "[]");
         break;
      }
      default:
         assert(0);
         break;
   }
   return descr;
} // dumpType

static void dumpRef(FILE* f, const char* className, const char* name, const char* descr)
{
   // neni-li to funkce, vypiseme pouze typ a jmeno
   if( *descr != '(' ) {
      dumpType(f, descr);
      fprintf(f, " ");
      if( className ) {
         dumpClassName(f, className);
         fprintf(f, ".");
      }
      fprintf(f, "%s", name);
      return;
   }

   // jmeno konstruktoru nebudeme vypisovat
   bool constr = strcmp(name, "<init>") == 0 || strcmp(name, "<clinit>") == 0;

   // neni-li to konstruktor, vypiseme typ vysledku
   if( !constr ) {      
      char* resultType = strchr(descr, ')');
      assert( resultType != NULL );
      dumpType(f, resultType + 1);
      fprintf(f, " ");
   }
   // pokud je zadano jmeno tridy, pridame jako prefix
   if( className ) {
      dumpClassName(f, className);
      if( !constr )
         fprintf(f, ".");
   }

   // neni-li to konstruktor, vypiseme jmeno metody
   if( !constr )
      fprintf(f, "%s", name);
   fprintf(f, "(");

   // a nakonec vypiseme parametry
   descr++;
   bool first = true;
   while( *descr != ')' ) {
      if( first )
         first = false;
      else
         fprintf(f, ",");
      descr = dumpType(f, descr);
   }
   fprintf(f, ")");
} // dumpRef

void JavaConstP::dump(FILE* f)
{
   static const char* constName[] = {
      "Unknown",    "Utf8",   "Unicode",  "Integer",  "Float",  "Long",
      "Double",     "Class",  "String",   "Field",    "Method", "InterfaceMethod",
      "NameAndType",
   };

   fprintf(f, "#%u <%s ", index, constName[tag]);
   switch( tag ) {
      case CONSTANT_Utf8:
         fprintf(f, "\"%s\"", s);
         break;
      case CONSTANT_Integer:
         fprintf(f, "%ld", i);
         break;
      case CONSTANT_Float:
         fprintf(f, "%#g", f);
         break;
# ifdef __MSC_VER
      case CONSTANT_Long:
         fprintf(f, "%I64i", l);
         break;
      case CONSTANT_Unicode:
         break;      
# endif
      case CONSTANT_Double:
         fprintf(f, "%#g", d);
         break;
      case CONSTANT_Class:
      case CONSTANT_String:
         fprintf(f, "\"%s\"", str->s);
         break;
      case CONSTANT_Fieldref:
      case CONSTANT_Methodref:
      case CONSTANT_InterfaceMethodref:
         dumpRef(f, ref.cls->str->s, ref.nt->nt.name->s, ref.nt->nt.type->s);
         break;
      case CONSTANT_NameAndType:
         dumpRef(f, NULL, nt.name->s, nt.type->s);
         break;
      default:
         fprintf(f, "???");
         break;
   }
   fprintf(f, ">");
} // JavaConstP::dump

bool JavaClassP::dump(const char* fileName, bool dumpCode, bool dumpConsts)
{
   int i;
   bool wideFlag = false;

   assert(thisClass != NULL);
   assert(superClass != NULL);

   // zjistime, do ktereho souboru se ma zapisovat, a otevreme ho pro zapis
   if( fileName == NULL ) {
      char newFileName[_MAX_PATH];
      sprintf(newFileName, "%s.dmp", thisClass->str->s);
      f= fopen(newFileName, "w");
   } else if( strcmp(fileName, "stdout") == 0 )
      f = stdout;
   else
      f = fopen(fileName, "w");
   
   if( f == NULL )
      return false;
   
   if( dumpConsts ) {
      fprintf(f, "Constant pool:\n");
      for(i = 1; i < constPool.getCount(); i++) {
         JavaConstP* cp = constPool[i];
         fprintf(f, "   ");
         cp->dump(f);
         fprintf(f, "\n");
      }
      fprintf(f, "\n");
   }

   if( sourceFile != NULL )
      fprintf(f, "Compiled from %s\n", sourceFile->s);

   if( JavaClass::ACC_PUBLIC   & accessFlags ) fprintf(f, "public ");
   if( JavaClass::ACC_FINAL    & accessFlags ) fprintf(f, "final ");
   if( JavaClass::ACC_ABSTRACT & accessFlags ) fprintf(f, "abstract ");
   fprintf(f, JavaClass::ACC_INTERFACE & accessFlags ? "interface " : "class ");
   dumpClassName(f, thisClass->str->s);
   fprintf(f, " extends ");
   dumpClassName(f, superClass->str->s);
   fprintf(f, "\n");

   if( JavaClass::ACC_SUPER & accessFlags )
      fprintf(f, "    /* ACC_SUPER bit set */\n");
   fprintf(f, "\n{\n");

   for(i = 0; i < fields.getCount(); i++) {
      JavaFieldP* fp = fields[i];
      fprintf(f, "   ");
      if( JavaClass::ACC_PUBLIC    & fp->accessFlags ) fprintf(f, "public ");
      if( JavaClass::ACC_PRIVATE   & fp->accessFlags ) fprintf(f, "private ");
      if( JavaClass::ACC_PROTECTED & fp->accessFlags ) fprintf(f, "protected ");
      if( JavaClass::ACC_STATIC    & fp->accessFlags ) fprintf(f, "static ");
      if( JavaClass::ACC_FINAL     & fp->accessFlags ) fprintf(f, "final ");
      if( JavaClass::ACC_VOLATILE  & fp->accessFlags ) fprintf(f, "volatile ");
      if( JavaClass::ACC_TRANSIENT & fp->accessFlags ) fprintf(f, "transient ");
      dumpRef(f, NULL, fp->name->s, fp->descriptor->s);
      fprintf(f, ";\n");
   }
   
   for(i = 0; i < methods.getCount(); i++) {
      JavaMethodP* mp = methods[i];
      fprintf(f, "   ");
      if( JavaClass::ACC_PUBLIC       & mp->accessFlags ) fprintf(f, "public ");
      if( JavaClass::ACC_PRIVATE      & mp->accessFlags ) fprintf(f, "private ");
      if( JavaClass::ACC_PROTECTED    & mp->accessFlags ) fprintf(f, "protected ");
      if( JavaClass::ACC_STATIC       & mp->accessFlags ) fprintf(f, "static ");
      if( JavaClass::ACC_FINAL        & mp->accessFlags ) fprintf(f, "final ");
      if( JavaClass::ACC_SYNCHRONIZED & mp->accessFlags ) fprintf(f, "synchronized ");
      if( JavaClass::ACC_NATIVE       & mp->accessFlags ) fprintf(f, "native ");
      if( JavaClass::ACC_ABSTRACT     & mp->accessFlags ) fprintf(f, "abstract ");
      if( strcmp(mp->name->s, "<init>") == 0 )  // je to konstruktor instance
         dumpRef(f, thisClass->str->s, mp->name->s, mp->descriptor->s);
      else
         dumpRef(f, NULL, mp->name->s, mp->descriptor->s);
      fprintf(f, ";\n");
   }
   fprintf(f, "}\n\n");

   if( dumpCode ) {
      for(i = 0; i < methods.getCount(); i++) {
         JavaMethodP* mp = methods[i];
         fprintf(f, "Method ");
         if( strcmp(mp->name->s, "<init>") == 0 )  // je to konstruktor instance
            dumpRef(f, thisClass->str->s, mp->name->s, mp->descriptor->s);
         else
            dumpRef(f, NULL, mp->name->s, mp->descriptor->s);
         fprintf(f, "\n");

         JavaCodeP* code = mp->code;
         if( code == NULL ) continue;

         fprintf(f, "     .limits %u, %u\n", code->maxStack, code->maxLocals);
         code->currentPC = 0;
         while( code->currentPC < code->codeSize ) {
            JavaUShort basePC = code->currentPC;
            JavaInstr oc = (JavaInstr)(JavaUByte)code->fetchByte();
            fprintf(f, "%4u %s ", basePC, instrName[oc]);
 
            switch( oc ) {
            case BIPUSH:
               fprintf(f, "%d", code->fetchByte());
               break;
            case SIPUSH:
               fprintf(f, "%d", code->fetchShort());
               break;
            case GOTO:
            case IF_ICMPEQ: case IF_ICMPNE: case IF_ICMPLT:
            case IF_ICMPGT: case IF_ICMPGE: case IF_ICMPLE:
            case IFEQ: case IFNE: case IFLT: case IFGT: case IFGE: case IFLE: {
               JavaShort off = code->fetchShort();
               fprintf(f, "%u", basePC + off);
               break;
            }
            case NEW: case ANEWARRAY:
            case LDC_W:
            case GETSTATIC: case PUTSTATIC: case GETFIELD: case PUTFIELD:
            case INVOKESTATIC: case INVOKEVIRTUAL: case INVOKESPECIAL: {
               JavaUShort indx = code->fetchShort();
               JavaConstP* cp = constPool[indx];
               cp->dump(f);
               break;
            }
            case LDC: {
               JavaUByte indx = code->fetchByte();
               JavaConstP* cp = constPool[indx];
               cp->dump(f);
               break;
            }
            case ALOAD: case ASTORE:
            case ILOAD: case ISTORE: {
               JavaUInt indx;
               if( wideFlag ) {
                  indx = code->fetchShort();
                  wideFlag = false;
               } else
                  indx = code->fetchByte();
               fprintf(f, "%u", indx);
               break;
            }

            case WIDE:
               wideFlag = true;
               break;

            case NEWARRAY: {
               JavaUByte indx = code->fetchByte();
               fprintf(f, "%s", typeName[indx-4]);
               break;
            }
            default:
               break;
            }
            fprintf(f, "\n");
         }
         fprintf(f, "\n");
      }
   }

   if( f != stdout )
      fclose(f);
   return true;
} // JavaClassP::dump


// CTENI CILOVEHO SOUBORU
// ======================

static void readData(void* ptr, size_t size)
{
   fread(ptr, 1, size, f);
} // readData

static void skipData(JavaUShort size)
{
   fseek(f, size, SEEK_CUR);
} // skipData

static unsigned char readUnsignedByte()
{
   unsigned char x;
   x = getc(f);
   return x;
} // readUnsignedByte

static unsigned short readUnsignedWord()
{
   unsigned char b[2];
   readData(b, 2);
   
   unsigned short x = ((unsigned short)b[0] << 8) | b[1];
   return x;
} // readUnsignedWord

static long readInt()
{
   unsigned char b[4];
   readData(b, 4);
   
   long x = (((((((long)b[0] << 8) | b[1]) << 8) | b[2])) << 8) | b[3];
   return x;
} // readInt

void JavaClassP::relocConst(JavaConstP* cp)
{
   switch( cp->tag ) {
   case JavaConst::CONSTANT_Class:
   case JavaConst::CONSTANT_String:
      cp->str = constPool[(JavaUShort)cp->str];
      break;
   case JavaConst::CONSTANT_Methodref:
   case JavaConst::CONSTANT_InterfaceMethodref:
   case JavaConst::CONSTANT_Fieldref:
      cp->ref.cls = constPool[(JavaUShort)cp->ref.cls];
      cp->ref.nt = constPool[(JavaUShort)cp->ref.nt];
      break;
   case JavaConst::CONSTANT_NameAndType:
      cp->nt.name = constPool[(JavaUShort)cp->nt.name];
      cp->nt.type = constPool[(JavaUShort)cp->nt.type];
      break;
   default:
      break;
   }
} // JavaClassP::relocConst

JavaConstP* JavaClassP::readConst()
{
   JavaConst::JavaConstTag tag = JavaConst::JavaConstTag(readUnsignedByte());
   
   JavaConstP* cp = new JavaConstP(tag);
   
   switch( tag ) {
   case JavaConst::CONSTANT_Utf8: {
      unsigned short len = readUnsignedWord();
      cp->s = new char[len+1];
      readData(cp->s, len);
      cp->s[len] = '\0';
      break;
   }
   case JavaConst::CONSTANT_Integer:
      cp->i = readInt();
      break;
   case JavaConst::CONSTANT_Class:
   case JavaConst::CONSTANT_String:
      cp->str = (JavaConstP*)readUnsignedWord();
      break;
   case JavaConst::CONSTANT_Methodref:
   case JavaConst::CONSTANT_InterfaceMethodref:
   case JavaConst::CONSTANT_Fieldref:
      cp->ref.cls = (JavaConstP*)readUnsignedWord();
      cp->ref.nt = (JavaConstP*)readUnsignedWord();
      break;
   case JavaConst::CONSTANT_NameAndType:
      cp->nt.name = (JavaConstP*)readUnsignedWord();
      cp->nt.type = (JavaConstP*)readUnsignedWord();
      break;
   default:
      assert(0);
      break;
   }
   return cp;
} // JavaClassP::readConst

JavaFieldP* JavaClassP::readField()
{
   JavaFieldP* fp = new JavaFieldP(this);
   fp->accessFlags = readUnsignedWord();
   fp->name = constPool[readUnsignedWord()];
   fp->descriptor = constPool[readUnsignedWord()];
   
   // atributy
   unsigned short numAttrs = readUnsignedWord();
   for(int i = 0; i < numAttrs; i++) {
      JavaConstP* attrTag = constPool[readUnsignedWord()];
      long attrLen = readInt();
      assert(attrTag->tag == JavaConst::CONSTANT_Utf8);
      if( strcmp(attrTag->s, "ConstantValue") == 0 ) {
         fp->constantValue = constPool[readUnsignedWord()];
      } else
         skipData((JavaUShort)attrLen);
   }
   
   return fp;
} // JavaClassP::readField

void JavaClassP::readCode(JavaMethodP* aMethod)
{
   JavaCodeP* cp = new JavaCodeP(aMethod);
   
   unsigned short maxStack = readUnsignedWord();
   unsigned short maxLocals = readUnsignedWord();
   cp->setLimits(maxStack, maxLocals);
   
   JavaUShort codeSize = (JavaUShort)readInt();
   JavaUByte* code = new JavaUByte[codeSize];
   readData(code, codeSize);
   cp->setCode(code, (unsigned short)codeSize);

   // tabulka vjimek
   unsigned short numExcps = readUnsignedWord();
   assert(numExcps == 0);
   
   // atributy
   unsigned short numAttrs = readUnsignedWord();
   for(int i = 0; i < numAttrs; i++) {
      JavaConstP* attrTag = constPool[readUnsignedWord()];
      long attrLen = readInt();
      assert(attrTag->tag == JavaConst::CONSTANT_Utf8);
      skipData((JavaUShort)attrLen);
   }
} // JavaClassP::readCode

JavaMethodP* JavaClassP::readMethod()
{
   JavaMethodP* mp = new JavaMethodP(this);
   
   mp->accessFlags = readUnsignedWord();
   mp->name = constPool[readUnsignedWord()];
   mp->descriptor = constPool[readUnsignedWord()];
   
   // atributy
   unsigned short numAttrs = readUnsignedWord();
   for(int i = 0; i < numAttrs; i++) {
      JavaConstP* attrTag = constPool[readUnsignedWord()];
      long attrLen = readInt();
      assert(attrTag->tag == JavaConst::CONSTANT_Utf8);
      if( strcmp(attrTag->s, "Code") == 0 ) {
         readCode(mp);
      } else
         skipData((JavaUShort)attrLen);
   }
   
   return mp;
} // JavaClassP::readMethod

JavaClassP* JavaClassP::load(const char* aThisClass)
{
   int i;
   char fileName[_MAX_PATH];
# ifdef __BORLANDC__
   sprintf(fileName, "%s.cla", aThisClass);
# else
   sprintf(fileName, "%s.class", aThisClass);
# endif

   f = fopen(fileName, "rb");
   if( f == NULL )
      return NULL;
   
   // zhlav
   if( readInt() != JavaClassMagicNumber ||
      readUnsignedWord() != JavaClassMinorVersion ||
      readUnsignedWord() != JavaClassMajorVersion )
   {
      fclose(f);
      return NULL;
   }
   
   JavaClassP* pp = new JavaClassP();
   
   // tabulka konstant
   unsigned short count = readUnsignedWord();
   pp->constPool.alloc(count);
   pp->addConst(NULL);
   
   for(i = 1; i < count; i++) {
      JavaConstP* con = pp->readConst();
      pp->addConst(con);
   }
   
   // relokace tabulky konstant (vzajmene odkazy mezi konstantami)
   for(i = 1; i < count; i++) {
      JavaConstP* con = pp->constPool[i];
      pp->relocConst(con);
   }
   
   // pstupov prva
   pp->accessFlags = readUnsignedWord();
   
   // aktuln a bzov tda
   pp->thisClass = pp->constPool[readUnsignedWord()];
   pp->superClass = pp->constPool[readUnsignedWord()];
   
   // seznam rozhran
   count = readUnsignedWord();
   pp->interfaces.alloc(count);
   for(i = 0; i < count; i++) {
      pp->interfaces.insertLast(pp->constPool[readUnsignedWord()]);
   }
   
   // seznam poloek
   count = readUnsignedWord();
   pp->fields.alloc(count);
   for(i = 0; i < count; i++)
      pp->fields.insertLast(pp->readField());
   
   // seznam metod
   count = readUnsignedWord();
   pp->methods.alloc(count);
   for(i = 0; i < count; i++)
      pp->methods.insertLast(pp->readMethod());
   
   // seznam atribut
   
   fclose(f);
   return pp;
} // JavaClassP::JavaClassP

static JavaShort swapShort(JavaShort x)
{
   return (JavaShort)((x << 8) | ((x >> 8) & 0xFF));
} // swapShort

void JavaCodeP::defLabel(JavaLabel label, JavaUShort offset)
{
   JavaLabelP* lp = labels[label];
   assert(lp->offset == JavaUndefLabel);
   lp->offset = offset;
   
   for(JavaUShort pc = lp->lastRef; pc != 0; ) {
      JavaUShort newPC = swapShort(*(JavaUShort*)&code[pc]);
      *(JavaUShort*)&code[pc] = swapShort(offset - (pc-1));
      pc = newPC;
   }
   
   lp->lastRef = 0;
} // JavaClassP::defLabel

JavaLabel JavaCodeP::newLabel()
{
   JavaLabelP* lp = new JavaLabelP();
   lp->offset = JavaUndefLabel;
   lp->lastRef = 0;
   labels.insertLast(lp);
   return labels.getCount() - 1;
} // JavaClassP::newLabel

void JavaCodeP::check()
{
   for(int i = 0; i < labels.getCount(); i++) {
      if( labels[i]->offset == JavaUndefLabel )
         fprintf(stderr, "JavaCode Error: Undefined label");
   }
   labels.destroyAll();
} // JavaCodeP::check


// JavaConstP - polozka tabulky konstant
// =====================================

JavaConstP::JavaConstP(JavaConstTag aTag): tag(aTag), index(0)
{
} // JavaConstP::JavaConstP

JavaConstP::~JavaConstP()
{
   switch( tag ) {
      case CONSTANT_Utf8:
         delete s;
         break;
      default:
         break;
   }
} // JavaConstP::~JavaConstP
