// Soubor: JavaClsP.h
// Popis:  Definice soukrome casti popisu tridy
// Autor:  benes@dcse.fee.vutbr.cz (Miroslav Benes)

# ifndef _JAVACLSP_H_
# define _JAVACLSP_H_

# ifdef _MSC_VER
# include <stdlib.h>
# endif
# include <stdio.h>

# include "javacls.h"
# include "jctable.h"

const JavaInt	JavaClassMagicNumber  = 0xCafeBabeL;
const JavaShort JavaClassMajorVersion = 45;
const JavaShort JavaClassMinorVersion = 3;

class JavaConstP: public JavaConst {
public:
   JavaConstTag	tag;
   unsigned	short	index;
   
   union {
      char*             s;       // CONSTANT_Utf8
      JavaInt           i;       // CONSTANT_Integer
      JavaFloat         f;       // CONSTANT_Float
# ifdef _MSC_VER
      JavaLong          l;       // CONSTANT_Long
      wchar_t*          us;      // CONSTANT_Unicode
# endif
      JavaDouble        d;       // CONSTANT_Double

      JavaConstP*       str;     // CONSTANT_Class, CONSTANT_String (ukazuje na CONSTANT_Utf8)
      struct {
         JavaConstP*    cls;     // -> CONSTANT_Class
         JavaConstP*    nt;      // -> CONSTANT_NameAndType
      } ref;                     // CONSTANT_Fieldref, Methodref, InterfaceMethodref
      struct {
         JavaConstP*    name;    // -> CONSTANT->Utf8
         JavaConstP*    type;    // -> CONSTANT->Utf8
      } nt;                      // CONSTANT_NameAndType
   };
   
   void dump(FILE* f);
   JavaConstP(JavaConstTag aTag);
   ~JavaConstP();
}; // JavaConstP

class JavaLabelP {
public:
   JavaUShort        offset;              // adresa nvt
   JavaUShort        lastRef;             // adresa poslednho odkazu
}; // JavaLabelP

class JavaMethodP;

class JavaCodeP {
public:
   JavaUShort maxStack;                   // poet vyuitch poloek zsobnku
   JavaUShort maxLocals;                  // poet poloek loklnch promnnch
   
   JavaUShort currentPC;                  // aktuln adresa do kdu
   JavaUShort codeSize;                   // aktuln velikost kdu
   JavaUShort maxCodeSize;                // velikost pidlen pamti pro kd
   JavaUByte* code;                       // kd
   bool       accessible;                 // nsledujc instrukce je dostupn 
   
   // ExceptionTable
   // LineNumberTable
   // LocalVariableTable
   
   JavaMethodP* method;                   // metoda, ke kter kd pat
   
   Table<JavaLabelP> labels;              // tabulka nvt
   
   // nastaven limit
   void setLimits(JavaUShort aMaxStack, JavaUShort aMaxLocals);
   
   // nastaven kdu
   void setCode(JavaUByte* buf, JavaUShort size);

   // vpoet dlky atributu
   JavaUInt getAttrLength() const;
   
   // generovn dat
   void genData(void* buf, JavaUShort size);
   void genByte(JavaUByte x);
   void genShort(JavaShort x);
   void genInt(JavaInt x);
# ifdef _MSC_VER
   void genLong(JavaLong x);
# endif
   
   // generovn skokov instrukce
   void genJump(JavaInstr op, JavaUShort label);
   
   // definice nvt
   void defLabel(JavaUShort label, JavaUShort offset);
   
   // vytvoen nvt
   JavaUShort newLabel();
   
   // kontrola vygenerovanho kdu
   void check();
   
   // zpis atributu Code
   void write();
   
   // konstruktor
   JavaCodeP(JavaMethodP* aMethod);

   void fetchData(void* ptr, size_t size);
   JavaByte fetchByte();
   JavaShort fetchShort();
   JavaInt fetchInt();
# ifdef _MSC_VER
   JavaLong fetchLong();
# endif

}; // JavaCode

class JavaFieldP: public JavaField {
public:
   JavaClassP* parent;               // tda, do kter poloka pat
   JavaConstP*	name;                 // jmno poloky (CONSTANT_Utf8)
   JavaUShort  accessFlags;
   JavaConstP*	descriptor;           // deskriptor (CONSTANT_Utf8)
   JavaConstP*	constantValue;        // poten hodnota (CONSTANT_?)

   JavaValue   currentValue;         // aktuln hodnota

   // konstruktor
   JavaFieldP(JavaClassP* aClass);
   JavaFieldP(JavaClassP* aClass, JavaConstP* aName, JavaConstP* aDescr, unsigned short flags);
}; // JavaFieldP

class JavaMethodP: public JavaMethod {
public:
   JavaClassP*      parent;        // tda, do kter metoda pat
   JavaConstP*	     name;          // jmno metody (CONSTANT_Utf8)
   JavaUShort       accessFlags;
   JavaConstP*	     descriptor;    // deskriptor (CONSTANT_Utf8)
   JavaCodeP*	     code;          // vygenerovan kd
   
   // konstruktor
   JavaMethodP(JavaClassP* aClass);
   JavaMethodP(JavaClassP* aClass, JavaConstP* aName, JavaConstP* aDescr, unsigned short flags);
}; // JavaMethodP

class JavaClassP {
   friend class JavaClass;
   
public:
   JavaConstP*	      thisClass;         // aktuln tda (CONSTANT_Class)
   JavaConstP*	      superClass;        // bzov tda (CONSTANT_Class)
   JavaUShort        accessFlags;       // atributy
   
   Table<JavaConstP>  constPool;        // tabulka konstant
   Table<JavaConstP>  interfaces;       // tabulka rozhran
   Table<JavaFieldP>  fields;	          // tabulka promnnch
   Table<JavaMethodP> methods;          // tabulka metod
   
   JavaConstP*	       sourceFile;       // jmno zdrojovho souboru (CONSTANT_Utf8)

   JavaConstP*        sourceFileAttr;   // jmno atributu SourceFile (CONSTANT_Utf8)
   JavaConstP*        constantValueAttr;// jmno attributu ConstantValue (CONSTANT_Utf8)
   JavaConstP*        codeAttr;         // jmno atributu Code (CONSTANT_Utf8)
   JavaConstP*        exceptionsAttr;   // jmno atributu Exceptions (CONSTANT_Utf8)
   
public:
   // konstruktor
   JavaClassP(const char* aThisClass, const char* aSuperClass, JavaUShort flags);
   
   // destruktor
   ~JavaClassP();
   
   // nastaven atributu SourceFile
   void setSourceFile(const char* s);
   
   // deklarace poloky tdy
   JavaFieldP* addField(const char* aName, const char* aDescr, JavaUShort flags);
   
   // vyhledn poloky tdy
   JavaField* getField(const char* aName, const char* aDescr);

   // deklarace metody
   JavaMethodP* addMethod(const char* aName, const char* aDescr, JavaUShort flags);
   
   // vyhledn metody
   JavaMethod* getMethod(const char* aName, const char* aDescr);

   // naten ze souboru
   static JavaClassP* load(const char* aThisClass);

   // zruseni souboru .class
   void abort();

   // uloen do souboru
   bool save();
   
   // ladic vpis
   bool dump(const char* fileName, bool dumpCode, bool dumpConsts);
   
   JavaConstP* createUtf8Const(const char* s);
   JavaConstP* createClassConst(const char* cls);
   JavaConstP* createIntegerConst(JavaInt val);
   JavaConstP* createStringConst(const char* s);
   JavaConstP* createNameAndTypeConst(const char* name, const char* type);
   JavaConstP* createRefConst(JavaConst::JavaConstTag tag, const char* cls, const char* meth, const char* descr);
   
private:
   // default konstruktor
   JavaClassP();
   
   void writeConst(JavaConstP* cp);
   void writeField(JavaFieldP* fp);
   void writeMethod(JavaMethodP* mp);
   void addConst(JavaConstP* cp);
   
   JavaConstP* readConst();
   JavaFieldP* readField();
   JavaMethodP* readMethod();
   void readCode(JavaMethodP* aMethod);
   void relocConst(JavaConstP* cp);
}; // JavaClassP

# endif
