//
// JavaVM.h
//

# ifdef __MSC_VER
# pragma once
# endif

# ifndef _JAVAVM_H_
# define _JAVAVM_H_

# include "javacls.h"
# include <stdio.h>

class JavaState {
public:
   JavaCode    code;
   JavaUShort  currentPC;
   JavaInt*    locals;
   JavaUShort  maxLocals;
   JavaInt*    stack;
   JavaUShort  maxStack;
   JavaInt     top;
   JavaState*  previous;

   JavaState(JavaMethod* method);
   ~JavaState();
   void save();
   void restore();

   void push(JavaInt val);
   JavaInt pop();

}; // JavaState

class JavaVM {
public:
   void run(JavaClass& mainClass, FILE* iFile = NULL, FILE* oFile = NULL);
   JavaVM();
private:
   JavaClass*  cls;
   JavaState*  state;
   JavaValue   result;
   bool        running;
   FILE*       inpFile;
   FILE*       outFile;
   void stdIoCall(JavaCString methodName, JavaCString methodDescr);
   void procFieldOps(JavaInstr op, JavaUShort indx);
   void procInvokeOps(JavaInstr op, JavaUShort indx);
   void procReturnOps(JavaInstr op);
   void procNewArray(JavaUByte t);
   void step();
   void pushState(JavaMethod* method);
   void popState();
}; // JavaVM

# endif
