%{
#ifndef _MALLOC_H
  #include <malloc.h>
#endif
#define YYSTYPE AtribVal
#include <stdarg.h>
#include <fstream.h>
#include <stream.h>
#include <FlexLexer.h>
#include "Vector.h"
#include "types.h"
#include "ValueType.h"
#include "PasNode.h"
#include "PascalLexer.h"
#include "AtribVal.h"
#include "VarTable.h"
#include "Compiler.h"
#include "StandardProc.h"

void yyerror(char *s);
char *className;
extern Compiler *compiler;
int yylex(){
  compiler->getLexer()->yylex();
};
int whileP=FALSE;
%}

%token <node> KS_ARRAY
%token <node> KS_BEGIN
%token <node> KS_BREAK
%token <node> KS_BOOLEAN
%token <node> KS_CHAR
%token <node> KS_DO
%token <node> KS_ELSE
%token <node> KS_END
%token <node> KS_FUNCTION
%token <node> KS_IF
%token <node> KS_INTEGER
%token <node> KS_OF
%token <node> KS_PROCEDURE
%token <node> KS_PROGRAM
%token <node> KS_RETURN
%token <node> KS_STRING
%token <node> KS_THEN
%token <node> KS_VAR
%token <node> KS_WHILE


%token <node> STREDNIK
%token <node> LZAVORKA
%token <node> PZAVORKA
%token <node> CARKA
%token <node> ROVNITKO
%token <node> ROZSAH
%token <node> DVOJTECKA
%token <node> TECKA
%token <operType> REL_OPER
%token <node> PRIRAZENI
%token <node> LINDEX
%token <node> PINDEX

%token <boolVal> BOOL_CONST
%token <name> IDENT
%token <strVal> STRING
%token <intVal> CISLO
%token <charVal> ZNAK
%left <operType> ADD_OPER MINUS_OPER
%left <operType> MULL_OPER 
%left NOT_OPER

%type <vector> seznam_identifikatoru blok_dekl_prom seznam_deklaraci 
%type <vector> dekl_promenne dekl_p_p deklarace skut_parametry seznam_param 
%type <vector> parametry parametr prikazy 
%type <node> s prog_hlavicka hlavicka_deklarace
%type <node> expr e f t g volani proc_name func_name proc_hlavicka
%type <valueType> zakl_typ typ typ_param 
%type <node> dekl_proc hodnota
%type <node> blok prikaz promenna 
%type <nameLine> ident
%type <intVal> zacatekWhile 
%start s

%%
s:	hlavicka_deklarace blok TECKA
		{$$ = $1;
		 ((ProgramNode *)$$)->setCode((SequenceNode *)$2);
                 compiler->setProgramNode((ProgramNode *)$$);
		 compiler->setLocalTable(NULL);
		}

hlavicka_deklarace: prog_hlavicka deklarace
		{$$ = $1;
                 ((ProgramNode *)$$)->setDeclarations(new DeclsNode($2));
                 compiler->setLocalTable(((ProgramNode *)$$)->getMainProc()
			->getLocalTable());
		}


prog_hlavicka: KS_PROGRAM ident STREDNIK
		{$$ = ProgramNode::construct($2);
                 compiler->globalTable()->addVar((ProgramNode *)$$);
		 compiler->globalTable()->addVar(((ProgramNode *)$$)->
			getMainProc());
                 compiler->globalTable()->addStandardMethods();
		 className=$2->name;
                }

deklarace: //e
		{$$ = new Vector();
		}
	| deklarace blok_dekl_prom
		{$$ = $1;
		 $$->add($2);
		 compiler->globalTable()->addVars($2);
		 delete $2;
		}
	| deklarace dekl_proc STREDNIK
		{$$ = $1;
		 $$->add($2);
		}
	| error		//Vice chyb v deklaracich obecne
		{$$ = new Vector();
		}

blok_dekl_prom: KS_VAR seznam_deklaraci STREDNIK
		{$$ = $2;
		}

seznam_deklaraci: seznam_deklaraci STREDNIK dekl_promenne
		{$$=$1;
		 $$->add($3);
                 delete $3;
		}
	| dekl_promenne
		{$$ = $1;
		}

dekl_promenne: seznam_identifikatoru DVOJTECKA typ
		{$$=$1;
		 TabEntry::makeEntries($$, $3);
		}
	| error		//Vice chyb v bloku za VAR
		{$$ = new Vector();
		}

seznam_identifikatoru: seznam_identifikatoru CARKA ident
		{$$ = $1;
		 $$->add(new NameLine($3->name,$3->line));
		}
	| ident
		{$$ = new Vector();
                 $$->add(new NameLine($1->name,$1->line));
		}

typ:	zakl_typ
		{$$ = new ValueType($1->typ);
		}
	| KS_ARRAY LINDEX CISLO ROZSAH CISLO PINDEX KS_OF zakl_typ
		{$$ = new ValueType(($8->typ)|ATYP,$3,$5);
		}

zakl_typ: KS_INTEGER
		{$$ = new ValueType(ITYP);
		}
	| KS_CHAR
		{$$ = new ValueType(CTYP);
		}
	| KS_BOOLEAN
		{$$ = new ValueType(BTYP);
		}
	| KS_STRING
		{$$ = new ValueType(STYP);
		}

proc_hlavicka: KS_PROCEDURE proc_name seznam_param STREDNIK
		{$$ = $2;
		 ((ProcedureNode *)$$)->setParameters(new ParametersNode ($3));
		}
	| KS_FUNCTION func_name seznam_param DVOJTECKA typ STREDNIK
		{$$ = $2;
		 ((ProcedureNode *)$$)->setParameters(new ParametersNode($3));
		 ((ProcedureNode *)$$)->setType($5);
                 ((ReturnNode*)(compiler->localTable()->elementAt(0)))
			->setType($5);
		}

dekl_proc: proc_hlavicka dekl_p_p blok
		{$$ = $1;
		 ((ProcedureNode *)$$)->setCode((SequenceNode *)$3);
		 ((ProcedureNode *)$$)->setLocalDecls(new DeclsNode($2));
                 compiler->setLocalTable(NULL);
		}

proc_name: ident 
		{compiler->setLocalTable(new VarTable());
                 $$=ProcedureNode::construct($1->name, compiler->localTable(), 
			FALSE, $1->line);
		 compiler->globalTable()->addVar((ProcedureNode*)$$);
		}

func_name: ident 
		{compiler->setLocalTable(new VarTable());
                 $$=ProcedureNode::construct($1->name, compiler->localTable(), 
			TRUE, $1->line);
		 compiler->globalTable()->addVar((ProcedureNode*)$$);
		}
                 
seznam_param: //e
		{$$ = new Vector();
		}
	| LZAVORKA parametry PZAVORKA
		{$$ = $2;
		 compiler->localTable()->addVars($2,TRUE);
		}
	| error
		{$$ = new Vector();
		}

parametry: parametry STREDNIK parametr 
		{$$=$1;
		 $$->add($3);
                 delete $3;
		}
	| parametr
		{$$ = $1;
		}

parametr: seznam_identifikatoru DVOJTECKA typ_param
		{$$=$1;
		 TabEntry::makeEntries($$, $3);
		}

typ_param: zakl_typ
		{$$ = new ValueType($1->typ);
		}
	| KS_ARRAY KS_OF zakl_typ
		{$$ = new ValueType(($3->typ)|ATYP,0,0);
		}

dekl_p_p: //e 
		{$$ = new Vector();
		}
	|dekl_p_p blok_dekl_prom
		{$$ = $1;
		 $$->add($2);
		 compiler->localTable()->addVars($2);
		 delete $2;
		}

blok: 	KS_BEGIN prikazy KS_END
		{$$ = new SequenceNode ($2);
		}

prikazy: prikazy STREDNIK prikaz
		{$$ = $1;
                 $$->add($3);
		}
	| prikaz
		{$$ = new Vector();
		 $$->add($1);
		}

prikaz: //e
		{$$ = new EmptyStmntNode();
		}
	| promenna PRIRAZENI expr
		{$$ = LetStmntNode::construct((LeftValueNode *)$1,
				(ExprNode *)$3);
		}
	| KS_RETURN
		{$$ = new ReturnStmntNode();
		}
	| KS_WHILE expr KS_DO zacatekWhile prikaz
		{$$ = WhileStmntNode::construct((ExprNode *)$2,
			(StatementNode *)$5);
		 whileP=$4;
		} 
	| blok
		{$$ = $1;
		}
	| KS_BREAK
		{$$ = BreakStmntNode::construct(whileP);
		}
	| volani
		{$$ = new CallStmntNode((CallNode *)$1);
		}
	| KS_IF expr KS_THEN prikaz 
		{$$ = IfStmntNode::construct((ExprNode *)$2,
			(StatementNode *)$4);
		}
	| KS_IF expr KS_THEN prikaz KS_ELSE prikaz
		{$$ = IfElseStmntNode::construct((ExprNode *)$2,
			(StatementNode *)$4, (StatementNode *)$6);
		}

zacatekWhile: //e
		{$$=whileP;
		 whileP=TRUE;
		}

promenna: ident LINDEX expr PINDEX 
		{$$ = TabEntry::constructLeftValue($1->name,(ExprNode *)$3);
		}
	| ident 
		{$$ = TabEntry::constructLeftValue($1->name,NULL);
		}


expr:	e REL_OPER e
		{$$ = BinaryNode::construct($2,(ExprNode*)$1,(ExprNode*)$3);
		}
	| e
		{$$ = $1;
		}

e:	 e ADD_OPER t
		{$$ = BinaryNode::construct($2,(ExprNode*)$1,(ExprNode*)$3);
		}
	| e MINUS_OPER t
		{$$ = BinaryNode::construct($2,(ExprNode*)$1,(ExprNode*)$3);
		}
	| t
		{$$ = $1;
		}

t:	t MULL_OPER f
		{$$ = BinaryNode::construct($2,(ExprNode*)$1,(ExprNode*)$3);
		}
	| f
		{$$ = $1;
		}

f:	MINUS_OPER g
		{$$ = UnaryNode::construct(MINUS,(ExprNode*)$2);
		}
	| NOT_OPER g
		{$$ = UnaryNode::construct(NOT,(ExprNode*)$2);
		}
	| g
		{$$ = $1;
		}

g:	hodnota
        	{$$ = $1;
		}    
	| CISLO 
		{$$ = new IntConstNode($1);
		}
	| ZNAK
		{$$ = new CharConstNode($1);
		}
	| STRING 
		{$$ = new StrConstNode($1);
		}
	| LZAVORKA expr PZAVORKA 
		{$$ = $2;
		}
	| BOOL_CONST 
		{$$ = new BoolConstNode($1);
		}

hodnota: ident LINDEX expr PINDEX 
		{$$ = TabEntry::constructRightValue($1->name,(ExprNode *)$3,
			new RealParametersNode(new Vector()));
		}

	| ident LZAVORKA skut_parametry PZAVORKA LINDEX expr PINDEX
		{$$ = TabEntry::constructRightValue($1->name,(ExprNode *)$6,
			new RealParametersNode($3));
		}

	| ident error
		{$$ = new RightValueNode(ETYP);
		}
 
	| ident LZAVORKA skut_parametry PZAVORKA
		{$$ = TabEntry::constructRightValue($1->name,NULL,
			new RealParametersNode($3));
		}
	| ident 
		{$$ = TabEntry::constructRightValue($1->name,NULL,
			new RealParametersNode(new Vector()));
		}

volani:	ident LZAVORKA skut_parametry PZAVORKA
		{$$ = CallNode::construct($1->name,new RealParametersNode($3));
		}
	| ident 
		{$$ = CallNode::construct($1->name,
			new RealParametersNode(new Vector()));
		}

skut_parametry: skut_parametry CARKA expr
		{$$=$1;
		 $$->add($3);
		}
	| expr
		{$$ = new Vector();
		 $$->add($1);
		}

ident: IDENT
		{$$ = new NameLine($1,compiler->getLexer()->lineno());
		}
	| error
		{$$ = new NameLine(NULL,compiler->getLexer()->lineno());
		}

%%
void yyerror(char *s){
  compiler->yyerror(s);
}

